/*
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                           *
 *                              S N A R K   1 4                              *
 *                                                                           *
 *                     A PICTURE RECONSTRUCTION PROGRAM                      *
 *                                                                           *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

 dist.cpp,v 1.2 2008/08/25 16:11:08 jklukowska Exp

 COMPARE THE RECONSTRUCTION WITH THE TEST PICTURE
 ***** INPUT VARIABLES *****
 A,B:     TWO NELEM*NELEM ARRAYS CONTAINING THE TWO PICTURES
 TO BE COMPARED.
 LINEF:    A LOGICAL VARIABLE WHICH DETERMINES IF THE RESIDUAL
 IS TO BE COMPUTED IN LINE OR STRIP MODE.
 LIST,WEIGHT: ARRAYS NECESSARY FOR RAY AND WRAY.
 FLAG:     INDICATOR TO CONTROL RESIDUAL COMPUTATION.
 FLAG.NE.1 CAUSES RESIDUAL TO BE COMPUTED.
 ***** OUTPUT VARIABLES *****
 AVE(J):      THE AVERAGE DENSITY IN REGION J OF PICTURE B.
 DIS(J):      THE TWO NORM DISTANCE IN REGION J BETWEEN A AND B.
 REL(J):      THE ONE NORM DISTANCE IN REGION J BETWEEN A AND B.
 VAR(J):      THE VARIANCE IN REGION J OF PICTURE B.
 STD(J):      THE STANDARD DEVIATION IN REGION J OF B (=SQRT(VAR)).
 RESID:    THE RESIDUAL- I.E. THE TWO NORM OF THE DIFFERENCE
 BETWEEN THE PSEUDO PROJECTION OF B AND THE GIVEN
 PROJECTION DATA FOR THE WHOLE PICTURE.

 ---  Modified 4/1/88 to enable evaluation of subregions, either
 through boundary selection by geomtric descriptions OR (OR BOTH)
 pixel selection through density windowing wrt to phantom.
 ---  Only pixels within the specified regions and density window
 will be used in computing the metrics
 ---  A switch ISW is added to indicate whether the phantom is passed through
 the first array, a(), or the second array, b(), so that density
 selection can always be referenced wrt the phantom
 */

#include <cstdlib>       
#include <cstdio>
#include <cmath>
#include <stdio.h>
#include <iostream>
#include <limits>
#include "math.h"
#include "blkdta.h"
#include "geom.h"
#include "region.h"
#include "anglst.h"
#include "pseudo.h"
#include "blob.h"
#include "consts.h"
#include "anglst.h"
#include "creacm.h"
#include "eval.h"
#include "DistanceMeasureWSQD.h"

void Eval_class::dist(
INTEGER isw,
REAL* a,
REAL* b,
BOOLEAN linef,
INTEGER* list,
REAL* weight,
INTEGER flag,
REAL* ave,
REAL* dis,
REAL* rel,
REAL* var,
REAL* std,
REAL* resid,
BOOLEAN fresid,
REAL* klds,
BOOLEAN fklds,
REAL* wsqd,
BOOLEAN fwsqd)
{
	REAL xd;
	REAL yd;
	REAL arg;

	REAL theta, sinth, costh;

	REAL raysum;
	REAL psum;
	INTEGER numb;
	REAL snorm;
	REAL rinc;
	REAL drecon;
	REAL dfantom;
	REAL dtmp;
	REAL divis;
	REAL t1;
	REAL t2;
	REAL sumx;
	REAL rel1;
	REAL sumx2;

	INTEGER npt1;
	INTEGER rarea;
	INTEGER i1;
	INTEGER idx0;
	INTEGER objtyp;
	REAL rdis;
	REAL wsqds;

	// The near0 parameter is added to allow resetting non-zero variances
	// (presumably arising from rounding error), computed variance < near0
	// will be set to zero

	REAL near0 = (REAL) 1e-5;

	//  Repeat for other specified regions until all are considered
	for (INTEGER ObjJ = 0; ObjJ < ObjN; ObjJ++)
	{

		t1 = Obj[ObjJ].t1;
		t2 = Obj[ObjJ].t2;
		sumx = 0.0;
		rel1 = 0.0;
		sumx2 = 0.0;
		rdis = 0;
		wsqds = 0.0;
		divis = 0.0;
		npt1 = 0;

		objtyp = Obj[ObjJ].objcod;

		// Specified region is the whole picture
		if (objtyp == 0)
		{

			// Test if pixel density (of phantom) is within selected range,
			// pixel outside density range are excluded
			for (int i = 0; i < GeoPar.area; i++)
			{

				if (isw == 0)
				{
					dtmp = a[i];
				}
				else
				{
					dtmp = b[i];
				}

				if ((dtmp >= t1) && (dtmp <= t2))
				{
					dfantom = a[i];
					drecon = b[i];
					sumx += drecon;
					sumx2 += SQR(drecon);
					drecon = (REAL) fabs(dfantom - drecon);
					rel1 += drecon;
					divis += drecon * drecon;
					npt1++;
				}
			}
		}
		else
		{

			// Call subroutine "subreg" to compute parameters of specified region.
			subreg(ObjJ);

			idx0 = (Region.iy1 - 2) * GeoPar.nelem;

			for (int j = Region.iy1; j <= Region.iy2; j++)
			{
				xd = Region.xd0;
				yd = Region.yd0;
				idx0 += GeoPar.nelem;

				// Test if current pixel is in specified region
				for (int i = Region.ix1; i <= Region.ix2; i++)
				{
					if (objtyp == 1)
					{
						// Rectangular subregion
						if ((fabs(xd) > Region.aplus)
								|| (fabs(yd) > Region.bplus))
							goto L222;
					}
					else
					{
						// Elliptical subregion
						arg = (xd * xd) / Region.asq + (yd * yd) / Region.bsq;
						if (arg > 1.0001)
							goto L222;
					}

					// Test if pixel density is within selected range
					i1 = idx0 + i;

					if (isw == 0)
					{
						dtmp = a[i1];
					}
					else
					{
						dtmp = b[i1];
					}

					if ((dtmp >= t1) && (dtmp <= t2))
					{
						dfantom = a[i1];
						drecon = b[i1];
						sumx += drecon;
						sumx2 += SQR(drecon); // sumx2 += SQR(drecon - ave[ObjJ]);
						drecon = (REAL) fabs(dfantom - drecon);
						rel1 += drecon;
						divis += drecon * drecon;
						npt1++;
					}

					L222: xd += Region.d1;
					yd -= Region.d2;
				}

				Region.xd0 -= Region.d2;
				Region.yd0 -= Region.d1;
			}
		}

		rel[ObjJ] = rel1;
		Obj[ObjJ].npt = npt1;
		rarea = npt1;
		ave[ObjJ] = sumx / rarea;
		dis[ObjJ] = (REAL) sqrt(divis / rarea);
		var[ObjJ] = MAX0((REAL) 0.0, sumx2 / rarea - SQR(ave[ObjJ]));

		// Reset very small non-zero variance (presumably due to rounding
		// error) to zero

		if (var[ObjJ] <= near0)
		{
			var[ObjJ] = 0;
		}

		std[ObjJ] = (REAL) sqrt(var[ObjJ]);
	}

	// ALLOW USER TO SKIP EXPENSIVE STATISTIC COMPUTATION
	if ((fresid || fklds || fwsqd) && flag != 1)
	{
		rdis = 0;
		REAL kldis = 0;
		bool kldserrorreported=false;
		for (int np = 0; np < GeoPar.prjnum; np++)
		{
			rinc = GeoPar.pinc;
			Anglst.getang(np, &theta, &sinth, &costh);

			if (GeoPar.vri)
			{
				rinc = GeoPar.pinc * MAX0((REAL) fabs(sinth), (REAL) fabs(costh));
			}

			for (int nr = GeoPar.fusray; nr <= GeoPar.lusray; nr++)
			{
				raysum = Anglst.prdta(np, nr);
				psum = pseudo(b, np, nr, list, weight, &numb, &snorm, linef, TRUE);

				if (numb != 0)
				{
					if (linef && !GeoPar.line) raysum /= rinc;
					if (!linef && GeoPar.line) raysum *= rinc;
					rdis += SQR(psum - raysum);

					// calculate kullback-leibler distance
					if (!fklds)
					{
						// if it wasn't requested, set it to nan
						kldis=nan("");
					}
					else if (raysum<0 && !kldserrorreported)
					{
						// report on non-applicability
						std::cout << "\n****Warning: Negative raysum in prjfil detected, KLDS is not applicable and will be nan!";
						kldis=nan("");
						kldserrorreported=true;
					}
					// if it was requested and no errors occured, perform calculation
					else if (!kldserrorreported && !isnan(kldis) && kldis!=std::numeric_limits<double>::infinity() && kldis!=std::numeric_limits<double>::max())
					{
						if (raysum<=Consts.zero && psum<=Consts.zero)
						{
							// do nothing
						}
						else if (raysum<=Consts.zero)
						{
							kldis += psum;
						}
						else if (psum<=Consts.zero)
						{
							// set to infinity
							if (std::numeric_limits<double>::has_infinity) kldis = std::numeric_limits<double>::infinity();
							else kldis = std::numeric_limits<double>::max();
						}
						else
						{
							if ((raysum / psum) > Consts.zero) kldis += (raysum * log(raysum / psum));
							else kldis += (raysum * log(Consts.zero));

							kldis += psum - raysum;
						}
					}
				}
			}
		}

		*resid = (REAL) sqrt(rdis);
		*klds = kldis;

		if (fwsqd) {
			DistanceMeasureWSQD* distanceMeasureWSQD = new DistanceMeasureWSQD();
			*wsqd = distanceMeasureWSQD->calculateDistance(b, list, weight);
			delete distanceMeasureWSQD;
		}
	}
	return;
}
