/*
 ***********************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/imagewise_roi.c $
 $LastChangedRevision: 80 $
 $Date: 2014-07-01 21:01:54 -0400 (Tue, 01 Jul 2014) $
 $Author: agulati $
 ***********************************************************

 Previously part of SuperSNARK
 */

#include <assert.h>

#include "malloc.h"
#include "experimenter.h"
#include "read_eval_phantom1.h"
#include "read_eval_recon1.h"
#include "hit_ratio.h"

/* -----------------------------imagewise_roi.c -------------------------------
 //jk 01/11/2008     new fom introduced based on Narayan & Herman paper
 
 This function computes theimagewise-roi.

 INPUTS:
 itr1 - array of length niters containing the iteration numbers for the
 first algorithm.
 itr2 - array of length niters containing the iteration numbers for the
 second algorithm.
 niters - number of iterations to be compared.
 pairs - array of length numpairs containing the indices of the structures
 which are paired.
 numpairs - number of pairs of structures.
 keywrd1 - string containing the keyword which defines the first algorithm.
 keywrd2 - string containing the keyword which defines the second algorithm.

 OUTPUTS:
 imagewise_roi1 - array of length niters containing the hit ratio for
 the first algorithm.
 imagewise_roi2 - array of length niters containing the hit ratio for
 the second algorithm.

 */

void imagewise_roi(int* itr1, int* itr2, int niters, int* pairs, int numpairs, char* keywrd1, char* keywrd2, double* imagewise_roi1, double* imagewise_roi2)
{
	double *phantom, *recon1, *recon2;
	int *regions, numstr, i, j, *strarea, *signal, *nonsignal;
	double sumnsph = 0, sumnsrecon1 = 0, sumnsrecon2 = 0, sumsph = 0,
			sumsrecon1 = 0, sumsrecon2 = 0, avnsph = 0, avnsrecon1 = 0,
			avnsrecon2 = 0, sdnsph = 0, sdnsrecon1 = 0, sdnsrecon2 = 0,
			goodness_ph = 0, goodness_recon1 = 0, goodness_recon2 = 0;

	/*********** compute goodness of the phantom ****************************************/

	//read in abnormality index for phantom structures
	read_eval_phantom1(&regions, &numstr, &phantom, &strarea); //regions is not used anywhere later on

	//allocate memmory to store indecis of structures that contain signal and
	//the ones that do not contain signal
	signal = (int*) malloc(numpairs * sizeof(int));
	nonsignal = (int*) malloc(numpairs * sizeof(int));

	//scan the phantom array to identify the structures that contain signal
	//ASSUMPTION: the higher density is associated with the signal.
	//the lower with non-signal structure
	//compute sumnsph and sumsph
	for (i = 0; i < numpairs * 2; i += 2)
	{
		if (phantom[pairs[i] - 1] > phantom[pairs[i + 1] - 1])
		{
			signal[i / 2] = pairs[i] - 1;
			nonsignal[i / 2] = pairs[i + 1] - 1;
			//assert(phantom[signal[i/2]] > phantom[nonsignal[i/2]]);
		}
		else
		{
			signal[i / 2] = pairs[i + 1] - 1;
			nonsignal[i / 2] = pairs[i] - 1;
			//assert(phantom[signal[i/2]] > phantom[nonsignal[i/2]]);
		}

		sumnsph += phantom[nonsignal[i / 2]];
		sumsph += phantom[signal[i / 2]];

	}

	//compute the average density of non-signal containing objects
	avnsph = sumnsph / numpairs;

	//compute standard deviation of density of non-signal containing objects
	for (i = 0; i < numpairs; i++)
	{
		sdnsph += (phantom[nonsignal[i]] - avnsph)
				* (phantom[nonsignal[i]] - avnsph);
	}

	sdnsph = sqrt(sdnsph);
	goodness_ph = (sumsph - sumnsph) / sdnsph;

	/*********** compute goodness of the reconstructions ********************************/

	//allocate memory to store abnormality indexes for
	//structures in the reconstructions
	recon1 = (double *) malloc(numstr * sizeof(double));
	recon2 = (double *) malloc(numstr * sizeof(double));

	for (i = 0; i < niters; i++)
	{
		/* read in abnormality indexes for structures in
		 the reconstructions. Initialize the arrays
		 which will store the structural accuracy */

		sumnsrecon1 = 0;
		sumnsrecon2 = 0;
		sumsrecon1 = 0;
		sumsrecon2 = 0;
		avnsrecon1 = 0;
		avnsrecon2 = 0;
		sdnsrecon1 = 0;
		sdnsrecon2 = 0;
		goodness_recon1 = 0;
		goodness_recon2 = 0;

		read_eval_recon1(&itr1[i], keywrd1, numstr, recon1);
		read_eval_recon1(&itr2[i], keywrd2, numstr, recon2);
		imagewise_roi1[i] = 0.0;
		imagewise_roi2[i] = 0.0;

		//compute sumnsrecon1 and sumsrecon1
		//and  sumnsrecon2 and sumsrecon2
		for (j = 0; j < numpairs; j++)
		{
			sumnsrecon1 += recon1[nonsignal[j]];
			sumsrecon1 += recon1[signal[j]];

			sumnsrecon2 += recon2[nonsignal[j]];
			sumsrecon2 += recon2[signal[j]];

		}

		//compute the average density of non-signal containing objects
		avnsrecon1 = sumnsrecon1 / numpairs;
		avnsrecon2 = sumnsrecon2 / numpairs;

		//compute standard deviation of density of non-signal containing objects
		for (j = 0; j < numpairs; j++)
		{
			sdnsrecon1 += (recon1[nonsignal[j]] - avnsrecon1)
					* (recon1[nonsignal[j]] - avnsrecon1);
			sdnsrecon2 += (recon2[nonsignal[j]] - avnsrecon2)
					* (recon2[nonsignal[j]] - avnsrecon2);
		}

		sdnsrecon1 = sqrt(sdnsrecon1);
		sdnsrecon2 = sqrt(sdnsrecon2);
		goodness_recon1 = ((sumsrecon1 - sumnsrecon1) / sdnsrecon1);
		goodness_recon2 = ((sumsrecon2 - sumnsrecon2) / sdnsrecon2);
		//output the value of the normalized FOM
		imagewise_roi1[i] = goodness_recon1 / goodness_ph;
		imagewise_roi2[i] = goodness_recon2 / goodness_ph;
	}

	free(phantom); /* free memory for next function call */
	free(signal);
	free(nonsignal);
	free(recon1);
	free(recon2);
	free(strarea);

}
