/*
 ***********************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/noise.cpp $
 $LastChangedRevision: 80 $
 $Date: 2014-07-01 21:01:54 -0400 (Tue, 01 Jul 2014) $
 $Author: agulati $
 ***********************************************************
 */

#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <DIGGauss.h>
#include <DIGPoisson.h>
#include "consts.h"
#include "geom.h"
#include "noise.h"
#include "spctrm.h"
#include "uiod.h"

// bug221 - moved noise code to this class - swr - 03/03/07

Noise_class NoisePar = Noise_class();

Noise_class::Noise_class()
{
	quanin = 0;
	quanmn = 1.0;
	quancm = 1.0;
	sctnfl = false;
	sctnpk = 0.0;
	sctnwd = 0.0;
	ultnfl = false;
	ultnmn = 0.0;
	ultnsd = 0.0;
	addnfl = false;
	addnmn = 0.0;
	addnsd = 0.0;
	oldnp = -1;
	sctcnt = NULL;
	sctwts = NULL;
	calray = NULL;
}

void Noise_class::resetNoise()
{
	// jklukowska, bug 263, in case the noise was set
	// previously, we need to reset it to initial values
	quanin = 0;
	quanmn = 1.0;
	quancm = 1.0;
	sctnfl = false;
	sctnpk = 0.0;
	sctnwd = 0.0;
	ultnfl = false;
	ultnmn = 0.0;
	ultnsd = 0.0;
	addnfl = false;
	addnmn = 0.0;
	addnsd = 0.0;
	oldnp = -1;
	sctcnt = NULL;
	sctwts = NULL;
	calray = NULL;
}

Noise_class::~Noise_class()
{
	if (sctcnt != NULL)
		delete[] sctcnt;
	if (sctwts != NULL)
		delete[] sctwts;
	if (calray != NULL)
		delete[] calray;
}

void Noise_class::setAdditiveNoise(REAL mean, REAL std)
{
	addnfl = true;
	addnmn = mean;
	addnsd = std;
}

REAL Noise_class::additiveNoise(REAL value)
{
	return addnfl ? value + Gauss(addnmn, addnsd) : value;
}

void Noise_class::setMultiplicativeNoise(REAL mean, REAL std)
{
	ultnfl = true;
	ultnmn = mean;
	ultnsd = std;
}

REAL Noise_class::multiplicativeNoise(REAL value)
{
	return ultnfl ? value * Gauss(ultnmn, ultnsd) : value;
}

void Noise_class::setScatter(REAL peak, REAL width)
{
	sctnfl = true;
	sctnpk = peak;
	sctnwd = width;
	// PRE-COMPUTE THE WEIGHTS FOR SCATTER CALCULATION
	// FIRST CONPUTE THE EXTENT OF SCATTER IN TERMS OF NUMBER OF RAYS
	// WE NEED SPACE = USRAYS LONG FOR CONVOLUTING

	sctcnt = new REAL[GeoPar.usrays];
	nsctr = (int) (width / GeoPar.pinc + 1.0);

	sctwts = new REAL[nsctr];
	for (INTEGER n = 0; n < nsctr; n++)
	{
		sctwts[n] = (REAL) (peak * (1.0 - n * GeoPar.pinc / width));
	}
	sctwts[0] += 1.0;
}

void Noise_class::scatterNoise(REAL* projection)
{
	if (sctnfl)
	{

		// CONVOLUTE WITH THE SCATTER FUNCTION

		for (INTEGER nr = 0; nr < GeoPar.usrays; nr++)
		{
			sctcnt[nr] = projection[nr];
		}

		for (INTEGER nr = 0; nr < GeoPar.usrays; nr++)
		{
			REAL sum = sctcnt[nr] * sctwts[0];
			REAL vsum = sctwts[0];

			for (INTEGER nrsct = 1; nrsct < nsctr; nrsct++)
			{
				INTEGER nshift = nrsct;

				// CONVOLVE CAREFULLY

				if ((nr - nshift) >= 0)
				{
					vsum += sctwts[nrsct];
					sum += sctcnt[nr - nshift] * sctwts[nrsct];
				}

				if ((nr + nshift) < GeoPar.usrays)
				{
					vsum += sctwts[nrsct];
					sum += sctcnt[nr + nshift] * sctwts[nrsct];
				}

			}
			projection[nr] = sum / vsum;
		}
	}

}

void Noise_class::setQuantumNoise(INTEGER type, REAL mean, REAL calibration)
{
	quanin = type;
	quanmn = mean;
	quancm = calibration;
	setQuantumNoise();
}

void Noise_class::setQuantumNoise()
{
	// GET AVERAGE BACKGROUND ABSORPTION
	back = 0.0;
	if (quanin == PET)
	{
		for (INTEGER l = 0; l < Spctrm.nergy; l++)
		{
			back += Spctrm.engwt[l] * Spctrm.backgr[l];
		}
	}
	else
	{
		for (INTEGER l = 0; l < Spctrm.nergy; l++)
		{
			back += (REAL) (Spctrm.engwt[l] * exp(-Spctrm.backgr[l]));
		}
	}
	c = back;
	// COMPUTE THE MEAN NUMBER OF PHOTONS EXITING FOR ACTUAL AND
	// CALIBRATION MEASUREMENTS
	aback = quanmn * back;
	if (quanin == PET)
		aback = back;
	cback = quancm * aback;

	// IF CALIBRATION TYPE 2 GET THE USRAYS CALIBRATION VALUES AND STORE
	// IN WORK AREA
	if (quanin == CONSTANT_RAY)
	{

		calray = new REAL[GeoPar.nrays];

		for (INTEGER nr = 0; nr < GeoPar.nrays; nr++)
		{
			calray[nr] = Gauss(cback, (REAL) sqrt(cback))
					/ Gauss(cback, (REAL) sqrt(cback));
#ifdef FFCOMPARE

			fprintf(output,"\n calray= %36.30f", calray[nr]);
#endif
		}
	}
	oldnp = -1;

}

REAL Noise_class::applyQuantumNoise(INTEGER np, INTEGER nr, REAL sum)
{
	if (quanin == NONOISE)
		return sum;
	if (quanin == PET)
	{
		if (quanmn >= 0)
			sum = (REAL) Poisson(sum);
	}
	else
	{
		//    if (quanin > NONOISE) sum = (REAL) Poisson(sum);
		if (sum < (REAL) 100.0)
			sum = (REAL) Poisson(sum);
		if (sum >= (REAL) 100.0)
			sum = Gauss(sum, (REAL) sqrt(sum));
	}

	// CHECK IF THE OBJECT IS OPAQUE FOR THIS RAY; TOO MUCH ABSORPTION

	if ((sum <= Consts.zero * aback) && (quanin != PET))
	{
		fprintf(output,
				"\n **** in projection %5i ray %5i no photons get through", np,
				nr);
		fprintf(output,
				"\n **** in order to do tomography, more photons are needed");
		fprintf(output, "\n **** program aborted\n");

		exit(-1);
	}
	return sum;
}

REAL Noise_class::raysum(REAL raysum)
{
	return quanin == PET ? raysum : exp(-raysum);
}

REAL Noise_class::computeAttenuation(INTEGER np, INTEGER nr, REAL ind)
{
	if (np != oldnp && quanin == CONSTANT_PROJECTION)
	{
		c = Gauss(cback, (REAL) sqrt(cback)) / Gauss(cback, (REAL) sqrt(cback));
		oldnp = np;
	}
	if (quanin == PET)
	{
		if (quanmn < 0)
		{
			ind += aback;
		}
		else
		{
			ind += Poisson(aback);
		}
	}
	else
	{
		REAL a = ind;
		if (quanin > NONOISE)
			a /= Gauss(aback, (REAL) sqrt(aback));
		if (quanin == CONSTANT_RAY)
			c = calray[nr];
		if (quanin == VARIABLE)
			c = Gauss(cback, (REAL) sqrt(cback)) / Gauss(cback, (REAL) sqrt(cback));
		ind = (REAL) log(c / a);
	}
	return ind;
}
