/*
 ***********************************************************
 $SNARK_Header: S N A R K  1 4 - A PICTURE RECONSTRUCTION PROGRAM $
 $HeadURL: svn://dig.cs.gc.cuny.edu/snark/trunk/src/snark/trm4.cpp $
 $LastChangedRevision: 131 $
 $Date: 2014-07-14 19:49:33 -0400 (Mon, 14 Jul 2014) $
 $Author: olangthaler $
 ***********************************************************

 Implementation of the indicator function I(x) used for MLEM-STOP

*/

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include "blkdta.h"
#include "uiod.h"
#include "trm4.h"
#include "anglst.h"
#include "wray.h"
#include "math.h"
#include "projfile.h"
#include "blob.h"
#include "infile.h"

using namespace std;

void trm4_class::Init()
{
	BOOLEAN eol;
	// read reporting command (disabled by default)
	INTEGER rprt = CHAR2INT('r','p','r','t');
	if (InFile.getwrd(FALSE, &eol, &rprt, 1) == rprt)
	{
		report = true;
		cout << "\n         reporting is enabled";
		cout << "\n         reporting file: RPRTmlst";

		skips = InFile.getnum(FALSE, &eol);
		if (skips>0)
		{
			cout << "\n         report skip factor " << skips;
		}
		else
		{
			cout << "\n         reporting on every iteration";
		}
	}
	else
	{
		report = false;
		cout << "\n         reporting is disabled";
	}
}

BOOLEAN trm4_class::Run(REAL* recon, INTEGER* list, REAL* weight, INTEGER iter)
{
	// numerator and denominator of the formula
	REAL numerator = 0;
	REAL denominator = 0;
	// temporary variables needed during calculation
	REAL sumaijxj = 0;
	// # projections
	static INTEGER prjnum;
	// # the bigger value of usrays or snrays
	static INTEGER nrays;
	// array for projection data
	static REAL* prjdata;
	// temporary variable for the count of near-simultaneous arrivals of photon pairs for the ith LOR
	REAL bi = 0;
	// variables for geometry data
	static INTEGER*** pixelnumbers;
	static REAL*** weights;
	static INTEGER** numpixels;
	REAL snorm = 0;

	// gather data which only needs to be gathered during the first iteration
	if (iter == 1)
	{
		prjnum = GeoPar.prjnum;
		nrays = GeoPar.nrays;

		// max # pixels that may be intersected by a ray
		INTEGER pixelcount = GeoPar.nelem * 2;

		prjdata = new REAL[prjnum * nrays];
		for (INTEGER i = 0; i < prjnum; i++)
		{
			ProjFile.ReadProj(i, &prjdata[i * nrays], nrays);
		}

		// allocate memory and read list of aij for each ray
		numpixels = new INTEGER*[prjnum];
		pixelnumbers = new INTEGER**[prjnum];
		weights = new REAL**[prjnum];
		for (INTEGER i = 0; i < prjnum; i++)
		{
			numpixels[i] = new INTEGER[nrays];
			pixelnumbers[i] = new INTEGER*[nrays];
			weights[i] = new REAL*[nrays];
			for (INTEGER j = 0; j < nrays; j++)
			{
				pixelnumbers[i][j] = new INTEGER[pixelcount];
				weights[i][j] = new REAL[pixelcount];

			    if(Blob.pix_basis)
			    {
			    	wray(i, j, pixelnumbers[i][j], weights[i][j], &numpixels[i][j], &snorm);
			    }
			    else
			    {
			        Blob.bwray(i, j, pixelnumbers[i][j], weights[i][j], &numpixels[i][j], &snorm);
			    }
			}
		}
	}

	// calculate numerator and denominator
	for (INTEGER i = 0; i < prjnum; i++)
	{
		for (INTEGER j = 0; j < nrays; j++)
		{
			sumaijxj = 0;
			for (INTEGER k = 0; k < numpixels[i][j]; k++)
			{
				sumaijxj += (weights[i][j][k]) * recon[pixelnumbers[i][j][k]];
			}
			denominator += sumaijxj;

			bi = prjdata[i * nrays + j];
			numerator += pow(bi - sumaijxj, 2);
		}
	}

	// calculate I(x) and check if <=1
	if (numerator / denominator <= 1)
	{
		// free all allocated memory
		delete[] (prjdata);
		for (INTEGER i = 0; i < prjnum; i++)
		{
			for (INTEGER j = 0; j < nrays; j++)
			{
				delete[] (pixelnumbers[i][j]);
				delete[] (weights[i][j]);
			}
			delete[] (numpixels[i]);
			delete[] (pixelnumbers[i]);
			delete[] (weights[i]);
		}
		delete[] (numpixels);
		delete[] (pixelnumbers);
		delete[] (weights);

		// additional output
		if (report)
		{
			cout << "\n         MLEM-STOP indicator function I(x)=" << numerator/denominator;

			FILE* RPRTfile;
			if (iter==1)
			{
				RPRTfile = fopen("RPRTmlst", "w");
				fprintf(RPRTfile,"MLEM-STOP indicator function I(x) reporting output");
				fprintf(RPRTfile,"\n    iter         I(x)");
				fprintf(RPRTfile,"\n %5i  %20.9f", iter, numerator/denominator);
			}
			else
			{
				RPRTfile = fopen("RPRTmlst", "a");
				fprintf(RPRTfile,"\n %5i  %20.9f", iter, numerator/denominator);
			}
			fclose(RPRTfile);
		}

		return TRUE;
	}
	else
	{
		// additional output
		if (report && (iter==1 || skips==0 || (iter%(INTEGER)skips)==0))
		{
			cout << "\n         MLEM-STOP indicator function I(x)=" << numerator/denominator;

			FILE* RPRTfile;
			if (iter==1)
			{
				RPRTfile = fopen("RPRTmlst", "w");
				fprintf(RPRTfile,"MLEM-STOP indicator function I(x) reporting output");
				fprintf(RPRTfile,"\n    iter         I(x)");
				fprintf(RPRTfile,"\n %5i  %20.9f", iter, numerator/denominator);
			}
			else
			{
				RPRTfile = fopen("RPRTmlst", "a");
				fprintf(RPRTfile,"\n %5i  %20.9f", iter, numerator/denominator);
			}
			fclose(RPRTfile);
		}

		return FALSE;
	}
}
