/** @file eval_helper.cpp
    @package snark14Display
    @author Bruno M. Carvalho and Deniz Sarioz
    licensed under (open-source) QPL v1.0
    which accompanies this distribution in the file QPL
 */

/** @name eval_helper.hpp
    @type C++ header file
    @date March 2004
    @package snark14Display
    @purpose to help snark14Display in the interpretation of eval files, broadly speaking.
    @author deniz, who realizes after 1200 lines that a lot of space and (programmer) time 
    could have been saved by enumerated-typing of global resolution measure field names
 *
    @description this header file contains the implementation of classes, starting with:
    phantomeas: phantom measurement structure
    globmeas  : global resolution measurement structure
    PBPmeas   : point-by-point error measurement structure
  (1-line measurements in eval: glorified structures with constructors that take "measline_t" (char*))
 *
 phantomeas   globmeas    PBPmeas
   |            |           |
   |     globmeaset   PBPmeaset (info from a single execution of a single algorithm in a call to EVALUATE)
   |        |          |
   |       eval_execution   (info from a single call to EVALUATE, that includes calls like "EVALUATE BOTH")
   |         |
   |         |
  eval_exec_set    (info from all EVALUATE calls of a snark14 run)
 *
  You can read "is member of" relationships if you go top down, very intuitive up to abbreviation.
 *
  Attempted improvements to Bruno's version (which was < 1.0):
  - no more segmentation faults due to eval
  - dynamic parsing and allocation for 'residual', an optional field of a globmeas, that can vary WITHIN a globmeaset
  - variable number of fields in a point-by-point measurement structure (PBPmeas)
  - support for multiple EVALUATE calls in snark14
  - extrema calculated dynamically, "recursively"
 *
  Notes:
  - the number of fields of all PBPmeas structures is uniquely determined by the snark14 produced eval file
    deniz thinks it is proportional to the logarithm of phantom/reconstruction images' number-of-pixels-on-side
  - phantomeas can be tied to eval_exec_set due to multiplicity, since there is exactly one phantom in each snark run,
       but it can occur with OR without residual, in the same.
 *
  TO DO:
  - plot phantomeas with globals.
 */

#include <qwidget.h>
#include <qpoint.h>
#include <qpixmap.h>
#include <qpainter.h>
#include <qstring.h>
#include <exception>
#include <stdexcept>
#include <cstdio>
#include <iostream>
#include <deque>
#include <vector>

#include "eval_helper.hpp"
#include "displaylines.hpp"
#include "chooseRanges.hpp"

#include "line_real_t.hpp"
#include "plot_t.hpp"

//  typedef char* measline_t;
//  typedef char* execname_t;

// global (resolution) measurements go into global measurement sets
// point-by-point resolution measurements go into PBP measurement sets
// an evaluation can have

/** each phantomeas corresponds to a line of phantom stats in the eval file */

class phantomeas {
    int region;
    int area;
    double average;
    double variance;
    double stddev;
    double* residualp;
    double* kullbackp;
public:

    /** phantomeas 1-arg constructor */
    phantomeas(measline_t line) {
        //0     625    0.3622                        0.1860    0.4312
        //1     4225   0.3574                        0.2795    0.5287 1216.8134
        double resid;
        double kullback;
        int numthingsread =
                std::sscanf(line, "%d %d %lg %lg %lg %lg %lg",
                &region, &area, &average, &variance, &stddev, &resid, &kullback);
        if (numthingsread < 5)
            throw std::invalid_argument("line in phantomeas constructor must have at least 5 numeric fields");
        if (numthingsread == 5) {
            residualp = 0;
            kullbackp = 0;
        } else if (numthingsread == 6) {
        	kullbackp = 0;
            residualp = new double;
            *residualp = resid;
        } else if (numthingsread == 7) {
            residualp = new double;
            *residualp = resid;
            kullbackp = new double;
            *kullbackp = kullback;
        } else throw std::logic_error("Something very interesting is wrong in phantomeas constructor");
    } // --phantomeas 1-arg ctor

    /** phantomeas ("deep") copy ctor */
    phantomeas(const phantomeas& p) {
        region = p.region;
        area = p.area;
        variance = p. variance;
        stddev = p.stddev;
        if (p.residualp) {
            residualp = new double;
            *residualp = *(p.residualp);
        } else residualp = 0;
        if (p.kullbackp) {
        	kullbackp = new double;
            *kullbackp = *(p.kullbackp);
        } else kullbackp = 0;
    } // --phantomeas::operator=

    /** phantomeas deep-copying assignment operator */
    phantomeas & operator=(const phantomeas& p) {
        std::cout << "in phantomeas deep assignment operator!" << std::endl;
        if (this == &p) return *this; // Gracefully handle self assignment
        region = p.region;
        area = p.area;
        variance = p. variance;
        stddev = p.stddev;
        if (p.residualp) {
            residualp = new double;
            *residualp = *(p.residualp);
        } else residualp = 0;
        if (p.kullbackp) {
        	kullbackp = new double;
            *kullbackp = *(p.kullbackp);
        } else kullbackp = 0;
        return *this;
    } // --phantomeas::operator=

    int getRegion() const {
        return region;
    }
    //  int getArea()        const { return area;           }

    double getAverage() const {
        return average;
    }

    double getVariance() const {
        return variance;
    }

    double getStddev() const {
        return stddev;
    }

    bool hasResidual() const {
        return (0 != residualp);
    }

    bool hasKullback() const {
        return (0 != kullbackp);
    }

    /** getResidual() should not be called without calling hasResidual() */
    double getResidual() { // defensive programming
        if (0 == residualp)
            throw std::logic_error("called phantomeas::getResidual() without calling hasResidual() where there is no residual");
        // else
        return (*residualp);
    }

    /** getKullback() should not be called without calling hasKullback() */
    double getKullback() { // defensive programming
        if (0 == kullbackp)
            throw std::logic_error("called phantomeas::getKullback() without calling hasKullback() where there is no KL distance");
        // else
        return (*kullbackp);
    }

    void setNewResidual(double res) {
        residualp = new double;
        *residualp = res;
    }

    void setNewKullback(double kld) {
        kullbackp = new double;
        *kullbackp = kld;
    }

    double get(globalYVar_t globalYVar) {
        switch (globalYVar) {
            case AVERAGE:
                return average;
                break;
            case VARIANCE:
                return variance;
                break;
            case STDDEV:
                return stddev;
                break;
            case RESIDUAL:
                if (hasResidual())
                    return getResidual();
            case KULLBACK:
                if (hasKullback())
                    return getKullback();
            default:
                // it is likely for this to be called with, say, distance, which phantom has none
                return -2.0;
                // negative value ought to be be ignored by callee
        }
    } // phantomeas::get()

    /** show routine for phantom, for debug  */
    void show() {
        std::cout << "> Phantom:" << std::endl;
        std::cout << "region average variance stddev";
        if (hasResidual()) std::cout << " residual";
        if (hasKullback()) std::cout << " KL distance";
        std::cout << std::endl;
        std::cout << region << " " << average << " " << variance << " " << stddev;
        if (hasResidual()) std::cout << " " << getResidual();
        if (hasKullback()) std::cout << " " << getKullback();
        std::cout << std::endl;
        std::cout << "< done showing phantom" << std::endl;
    } // --show

    ~phantomeas() {
        //    std::cout << "in phantomeas destructor!" << std::endl;
        if (residualp) delete residualp;
        if (kullbackp) delete kullbackp;
    }
}; // --class phantomeas

/** a globmeas corresponds to one line of global resolution stats in the eval file */
class globmeas {
    // algname now in globmeaset
    int iter; // some iterations might be missing from eval file => not always 1..n
    int area;
    double average;
    double distance;
    double relerr;
    double variance;
    double stddev;
    double* residualp; // uninitialized if no residual
    double* kullbackp; // uninitialized if no kl distance
    static unsigned NEXT_GMID;
    unsigned GMID;
    // setfields(int iter, int area, double average, double distance
public:

    globmeas(measline_t line) {
        GMID = NEXT_GMID++;
        //    1      4225    0.3850    0.8711    0.8927    0.0122    0.1104 9756.9313
        //  char *line2 = "  1      4225    0.3850    0.8711    0.8927    0.0122    0.1104 9756.9313";
        //  char *line2 = "  1      4225    0.3850    0.8711    0.8927    0.0122    0.1104          ";
        double resid;
        double kldis;
        int numthingsread =
                std::sscanf(line, "%d %d %lg %lg %lg %lg %lg %lg %lg",
                &iter, &area, &average, &distance, &relerr, &variance, &stddev, &resid, &kldis);
        if (numthingsread < 7)
            throw std::invalid_argument("line in globmeas constructor must have at least 7 numeric fields");
        if (numthingsread == 7) {
            residualp = 0;
            kullbackp = 0;
        } else if (numthingsread == 8) {
            residualp = new double;
            *residualp = resid;
            kullbackp = 0;
        } else if (numthingsread == 9) {
            residualp = new double;
            *residualp = resid;
            kullbackp = new double;
            *kullbackp = kldis;
        } else throw std::logic_error("Something very interesting is wrong in globmeas constructor");
    } // --globmeas 1-arg ctor

    /** globmeas deep copy constructor */
    globmeas(const globmeas& g) {
        if (std::verbose >= 5) std::cout << "In deep copy constructor with rhs GMID " << g.getGMID() << std::endl;
        GMID = g.GMID;
        iter = g.iter;
        area = g.area;
        average = g.average;
        distance = g.distance;
        relerr = g.relerr;
        variance = g. variance;
        stddev = g.stddev;
        if (g.residualp) {
            residualp = new double;
            *residualp = *(g.residualp);
        } else residualp = 0;
        if (g.kullbackp) {
        	kullbackp = new double;
            *kullbackp = *(g.kullbackp);
        } else kullbackp = 0;
    } // --globmeas copy ctor

    unsigned getGMID() const {
        return static_cast<unsigned> (GMID);
    }

    /** globmeas deep-copying assignment operator */
    globmeas & operator=(const globmeas& g) {
        if (std::verbose >= 5) std::cout << "In globmeas operator= with rhs GMID " << g.getGMID() << std::endl;
        if (this == &g) return *this; // Gracefully handle self assignment
        GMID = g.GMID;
        iter = g.iter;
        area = g.area;
        average = g.average;
        distance = g.distance;
        relerr = g.relerr;
        variance = g. variance;
        stddev = g.stddev;
        if (g.residualp) {
            residualp = new double;
            *residualp = *(g.residualp);
        } else residualp = 0;
        if (g.kullbackp) {
        	kullbackp = new double;
            *kullbackp = *(g.kullbackp);
        } else kullbackp = 0;
        return *this;
    } // --globmeas::operator=

    /** globmeas functions that 'get' fields */
    int getIter() const {
        return iter;
    }
    //  int    getArea()     const { return area; }

    double getAverage() const {
        return average;
    }

    double getDistance() const {
        return distance;
    }

    double getRelerr() const {
        return relerr;
    }

    double getVariance() const {
        return variance;
    }

    double getStddev() const {
        return stddev;
    }

    bool hasResidual() const {
        return (0 != residualp);
    }

    bool hasKullback() const {
        return (0 != kullbackp);
    }

    double getResidual() const {
        if (0 == residualp)
            throw std::logic_error("called getResidual without calling hasResidual() where there is no residual");
        // else
        return (*residualp);
    }

    double getKullback() const {
        if (0 == kullbackp)
            throw std::logic_error("called getKullback without calling hasKullback() where there is no kl distance");
        // else
        return (*kullbackp);
    }

    /** globmeas::show_all() is a debug routine to show all fields of a single global measurement of an iteration */
    void show_all() {
        std::cout << iter;
        std::cout << " ";
        std::cout << area;
        std::cout << " ";
        std::cout << average;
        std::cout << " ";
        std::cout << distance;
        std::cout << " ";
        std::cout << relerr;
        std::cout << " ";
        std::cout << variance;
        std::cout << " ";
        std::cout << stddev;
        if (hasResidual()) {
            std::cout << " ";
            std::cout << getResidual();
        }
        if (hasKullback()) {
            std::cout << " ";
            std::cout << getKullback();
        }
        std::cout << std::endl;
    } // globmeas::show_all()

    ~globmeas() {
        if (0 != residualp) {
            if (std::verbose >= 5) std::cout << "Residualp was nonzero, residual " << (*residualp) << " ";
            delete residualp;
        }
        if (0 != kullbackp) {
            if (std::verbose >= 5) std::cout << "Kullbackp was nonzero, kl distance " << (*kullbackp) << " ";
            delete kullbackp;
        }
        if (std::verbose >= 5) std::cout << "In globmeas " << GMID << " Dtor" << std::endl;
    } // --globmeas dtor
}; // --class globmeas

// unsigned globmeas::NEXT_GMID = 0; // class variable initialization

/** a globmeaset is the set of global resolution statistics for one execution of an algorithm */
class globmeaset {
    execname_t execname;
    //  QString execname;   // name of the execution
    // note that the same algorithm can have several executions
    char algname[5]; // name of the algorithm (3 or 4 letters followed by null terminator)
    std::deque<globmeas> measurements;
    // why a deque and not a vector?
    // --all access is sequential
    // --regular queue does not allow iterator (i think)
    // --when you push_back something new, it does not destruct all existing elements, ensuring linear runtime
    // --can be used a lot like a vector (.size() etc are available)
    double minAverage, maxAverage, minDistance, maxDistance, minRelerr, maxRelerr;
    double minVariance, maxVariance, minStddev, maxStddev;
    double minResidual, maxResidual; // both -1 if no line with residual in the set
    double minKullback, maxKullback; // both -1 if no line with kl distance in the set
    //  int minArea, maxArea;
    int minIter, maxIter; // minIter is not necessarily 1
    static unsigned NEXT_GMSID; // one for all global measurement sets
    unsigned GMSID; // unique ID for this set
    bool completed;
    bool extrema_calculated;
    char name[512];
public:

    globmeaset(execname_t line) {
        GMSID = NEXT_GMSID++;
        completed = false;
        extrema_calculated = false;
        memset(name, 0, 512);
        char *p = line;
        char *q;
        q = &(name[0]);

        static const char *needle = "xecution name:";
        p = strstr(p, needle);
        if (p) {
            p += strlen(needle);
            while ((' ' == *p) || ('\t' == *p)) p++;
            for (int i = 0; (i < 490) && (*p) && ('\n' != *p) && ('\r' != *p); i++) {
                *q++ = *p++;
            }
        }
        *q = '\0'; // just to be sure
        ///    std::cout << "exec name in globmeaset is: `" << name << "'" << std::endl;
        if (std::verbose >= 4) std::cout << "globmeaset " << GMSID << " ctor" << std::endl;
    } // --globmeaset ctor

    unsigned getGMSID() const {
        return GMSID;
    }

    execname_t getName() {
        return (&name[0]);
    }

    /*    void addMeasurement(globmeas& g) {    measurements.push_back(g);    }  */

    /** globmeaset::addMeasurement() */
    void addMeasurement(measline_t line) {
        if (completed)
            throw std::logic_error("Must NOT add to globmeaset once complete() is signalled");
        // else
        if (std::verbose >= 6) std::cout << "++ globmeaset::addMeasurement(line)" << std::endl;
        globmeas g(line);
        if (std::verbose >= 6) std::cout << "   globmeaset::addMeasurement(line): before push" << std::endl;
        measurements.push_back(g); // pass by reference, and a copy is made using copy constructor
        if (std::verbose >= 6) std::cout << "-- globmeaset::addMeasurement(line)" << std::endl;
    } // --globmeaset::addMeasurement(line)

    unsigned get_num_measurements() const {
        return (static_cast<unsigned> (measurements.size()));
    }

    void addGlobVarLine(line_real_set_t& LRS, globalYVar_t globalYVar) {
        line_real_t LR(getName());
        for (std::deque<globmeas>::iterator it = measurements.begin();
                it != measurements.end(); it++) {
            switch (globalYVar) {
                case AVERAGE:
                    LR.add(it->getIter(), it->getAverage());
                    break;
                case DISTANCE:
                    LR.add(it->getIter(), it->getDistance());
                    break;
                case RELERR:
                    LR.add(it->getIter(), it->getRelerr());
                    break;
                case VARIANCE:
                    LR.add(it->getIter(), it->getVariance());
                    break;
                case STDDEV:
                    LR.add(it->getIter(), it->getStddev());
                    break;
                case RESIDUAL:
                    if (it->hasResidual()) {
                        LR.add(it->getIter(), it->getResidual());
                    }
                case KULLBACK:
                    if (it->hasKullback()) {
                        LR.add(it->getIter(), it->getKullback());
                    }
                    break;
                default:
                    std::cerr << "Invalid globalYVar in globmeaset::addGlobVarLine()" << std::endl;
            } // switch
        } // for
        if (!LR.empty()) LRS.add(LR);
    } // --globmeaset::addGlobVarLine()

    /** globmeaset::complete() invoke to indicate that all elements have been added to this globmeaset */
    void complete() {
        if (std::verbose >= 4)
            std::cout << "Executing globmeaset::complete()" << std::endl;
        if (measurements.empty()) {
            throw std::logic_error("empty globmeaset refused to complete() itself");
        } // else
        if (completed)
            throw std::logic_error("complete() already signalled for this globmeaset, must NOT signal again.");
        // else
        if (std::verbose >= 5)
            std::cout << "completed is so far false" << std::endl;
        completed = true;
        if (std::verbose >= 5)
            std::cout << "completed is now set to true" << std::endl;
    } // globmeaset::complete()

    /** globmeaset::isComplete() whether or not this global measurement set has been completed */
    bool isComplete() const {
        return completed;
    } // globmeaset::isComplete()

    /** globmeaset::calculate_extrema() calculates extrema of all measurments by field
        assumes that 'iter' field is strictly increasing in the order they were added.
        should be called as either 0-arg or 3-arg                                       */
    bool calculate_extrema(bool restrict_domain = false, int min_it_arg = 0, int max_it_arg = 0) {
        if (!completed)
            throw std::logic_error("Must NOT calculate extrema of globmeaset before complete() is signalled");
        // else
        if (0 == get_num_measurements())
            throw std::logic_error("Must NOT calculate extrema of empty globmeaset"); // else
        //    int curArea;
        double curAverage, curDistance, curRelerr, curVariance, curStddev, curResidual, curKullback;
        minAverage = maxAverage = minDistance = maxDistance = minRelerr = maxRelerr = -1.0;
        minVariance = maxVariance = minStddev = maxStddev = minResidual = maxResidual = minKullback = maxKullback = -1.0;
        extrema_calculated = true; // this is intentionally true even if return value is false

        std::deque<globmeas>::const_iterator it = measurements.begin();

        if (restrict_domain) {
            if (min_it_arg > max_it_arg)
                throw std::invalid_argument("globmeaset::calculate_extrema() called with arguments min>max");
            while ((it != measurements.end()) && (min_it_arg > (*it).getIter()))
                it++;
            if (it == measurements.end())
                return false; // iterations of this globmeaset are out of requested range
            maxIter = minIter = (*it).getIter();
        } else {
            minIter = (*it).getIter();
            maxIter = measurements.back().getIter(); // and that's that.
        }
        //    curArea     = minArea     = maxArea     = (*it).getArea();
        curAverage = minAverage = maxAverage = (*it).getAverage();
        curDistance = minDistance = maxDistance = (*it).getDistance();
        curRelerr = minRelerr = maxRelerr = (*it).getRelerr();
        curVariance = minVariance = maxVariance = (*it).getVariance();
        curStddev = minStddev = maxStddev = (*it).getStddev();
        if ((*it).hasResidual()) { // not necessarily true
            curResidual = minResidual = maxResidual = (*it).getResidual();
        }
        if ((*it).hasKullback()) { // not necessarily true
            curKullback = minKullback = maxKullback = (*it).getKullback();
        }
        it++;
        for (; it != measurements.end(); it++) {
            if (restrict_domain) {
                if (max_it_arg < (*it).getIter())
                    break;
                else maxIter = (*it).getIter();
            }
            //      curArea     = (*it).getArea();
            curAverage = (*it).getAverage();
            curDistance = (*it).getDistance();
            curRelerr = (*it).getRelerr();
            curVariance = (*it).getVariance();
            curStddev = (*it).getStddev();
            //      if(curArea     < minArea)     minArea     = curArea;
            if (curAverage < minAverage) minAverage = curAverage;
            if (curDistance < minDistance) minDistance = curDistance;
            if (curRelerr < minRelerr) minRelerr = curRelerr;
            if (curVariance < minVariance) minVariance = curVariance;
            if (curStddev < minStddev) minStddev = curStddev;

            //      if(curArea     > maxArea)     maxArea     = curArea;
            if (curAverage > maxAverage) maxAverage = curAverage;
            if (curDistance > maxDistance) maxDistance = curDistance;
            if (curRelerr > maxRelerr) maxRelerr = curRelerr;
            if (curVariance > maxVariance) maxVariance = curVariance;
            if (curStddev > maxStddev) maxStddev = curStddev;

            if ((*it).hasResidual()) {
                curResidual = (*it).getResidual();
                if (minResidual < 0) { // this is the first residual in measurements
                    minResidual = maxResidual = curResidual;
                } else {
                    if (curResidual < minResidual) minResidual = curResidual;
                    if (curResidual > maxResidual) maxResidual = curResidual;
                }
            }
            if ((*it).hasKullback()) {
            	curKullback = (*it).getKullback();
                if (minKullback < 0) { // this is the first kl distance in measurements
                    minKullback = maxKullback = curKullback;
                } else {
                    if (curKullback < minKullback) minKullback = curKullback;
                    if (curKullback > maxKullback) maxKullback = curKullback;
                }
            }
        } // --foreach measurement
        return true;
    } // --globmeaset::calculate_extrema()

    /** these functions return appropriate domain extrema from a single execution of an algorithm */
    int getMinIter() {
        if (!extrema_calculated)
            calculate_extrema();
        return minIter;
    } // globmeaset::getMinIter()

    int getMaxIter() {
        if (!extrema_calculated)
            calculate_extrema();
        return maxIter;
    } // globmeaset::getMaxIter()

    /** these functions return appropriate range extrema from a single execution of an algorithm
        All will calculate_extrema() with no-args (entire domain) if
        it has never been called (with any set of arguments, and for a valid set)
     */
    double getMinAverage() {
        if (!extrema_calculated)
            calculate_extrema();
        return minAverage;
    } // globmeaset::getMinAvarage()

    double getMaxAverage() {
        if (!extrema_calculated)
            calculate_extrema();
        return maxAverage;
    }

    double getMinDistance() {
        if (extrema_calculated || calculate_extrema())
            return minDistance;
        else return -1.0;
    }

    double getMaxDistance() {
        if (extrema_calculated || calculate_extrema())
            return maxDistance;
        else return -1.0;
    }

    double getMinRelerr() {
        if (extrema_calculated || calculate_extrema())
            return minRelerr;
        else return -1.0;
    }

    double getMaxRelerr() {
        if (extrema_calculated || calculate_extrema())
            return maxRelerr;
        else return -1.0;
    }

    double getMinVariance() {
        if (extrema_calculated || calculate_extrema())
            return minVariance;
        else return -1.0;
    }

    double getMaxVariance() {
        if (extrema_calculated || calculate_extrema())
            return maxVariance;
        else return -1.0;
    }

    double getMinStddev() {
        if (extrema_calculated || calculate_extrema())
            return minStddev;
        else return -1.0;
    }

    double getMaxStddev() {
        if (extrema_calculated || calculate_extrema())
            return maxStddev;
        else return -1.0;
    }

    /* both of these return -1.0 if no residual */
    double getMinResidual() {
        if (extrema_calculated || calculate_extrema())
            return minResidual;
        else return -1.0;
    }

    double getMinKullback() {
        if (extrema_calculated || calculate_extrema())
            return minKullback;
        else return -1.0;
    }

    double getMaxResidual() {
        if (extrema_calculated || calculate_extrema())
            return maxResidual;
        else return -1.0;
    }

    double getMaxKullback() {
        if (extrema_calculated || calculate_extrema())
            return maxKullback;
        else return -1.0;
    }

    bool hasResidual() { // hack: think about whether this actually works
        if (extrema_calculated || calculate_extrema())
            return (-0.1 < minResidual);
        else return false;
    }

    bool hasKullback() { // hack: think about whether this actually works
        if (extrema_calculated || calculate_extrema())
            return (-0.1 < minKullback);
        else return false;
    }

    /** globmeaset::show_all() shows all members of this global measurement set */
    void show_all() {
        if (!completed)
            throw std::logic_error("Must NOT show_all globmeaset before complete() is signalled");
        // else
        for (std::deque<globmeas>::iterator mit = measurements.begin();
                mit != measurements.end(); mit++) {
            mit->show_all();
        }
    } // --globmeaset::show_all()

    ~globmeaset() {
        if (std::verbose >= 4)
            std::cout << "globmeaset " << GMSID << " Dtor" << std::endl;
    }
}; // --class globmeaset

// unsigned globmeas::NEXT_GMID = 0; // class variable initialization

/** a PBPmeas corresponds to one line of point-by-point statistics in the eval file */
class PBPmeas {
    // the fields are such that there is no reason for deep copies:
    //   default assignment and copy constructors are fine.
    // algname now in PBPmeaset
    int iter;
    std::vector<double> e; // it's imperative that this is something variable
    unsigned PMID;
    static unsigned NEXT_PMID;
public:
    // pre: line starts with whitespace after algname
    // pre: line points to a (non-static) buffer

    PBPmeas(measline_t line) {
        // read line into iter and the e vector:
        ///  if(std::verbose>=5) std::cout << "In PBPmeas " << (PMID+1) << " ctor" << std::endl;
        double myDouble;
        char *p;
        if (std::verbose >= 6) std::cout << "++ PBPmeas ctor(line)" << std::endl;
        //    std::cout << "PBPmeas beginning--Line is:" << std::endl << line << std::endl;
        p = line;
        while (((*p) == ' ') || ((*p) == '\t')) p++; // death to strtok!
        if (1 != std::sscanf(p, "%d", &iter))
            throw std::invalid_argument("line in PBPmeas ctor must start with \
integer iter");
        while ((*p) && ((*p) != ' ') && ((*p) != '\t')) p++; // skip iter
        while (((*p) == ' ') || ((*p) == '\t')) p++; // skip whitespace after iter
        while ((*p) && ((*p) != '\n') && ((*p) != '\r')) { // read in each double (might have to find out EOF symbol)
            if (1 == std::sscanf(p, "%lg", &myDouble)) {
                e.push_back(myDouble);
            }
            while ((*p) && ((*p) != ' ') && ((*p) != '\t')) p++; // skip double
            while (((*p) == ' ') || ((*p) == '\t')) p++; // skip whitespace after double
        }
        if (e.empty()) {
            std::cout << "e is empty... Line is:" << std::endl << line << std::endl;
            throw std::invalid_argument("line arg to PBPmeas ctor must have \
at least 1 floating point measurement");
        }
        PMID = NEXT_PMID++;
        if (std::verbose >= 6) std::cout << "-- PBPmeas ctor(line)" << std::endl;
    } // --PBPmeas ctor

    int getIter() const {
        return iter;
    }

    unsigned get_num_Es() const {
        return ( static_cast<unsigned> (e.size()));
    } // --PBPmeas::get_num_Es()

    double get_E(unsigned ind) {
        if (ind >= e.size())
            throw std::invalid_argument("Bad index to PBPmeas::get_E()");
        // else
        return e[ind];
    } // PBPmeas::get_E()

    void show_all() {
        std::cout << getIter() << " ";
        for (std::vector<double>::const_iterator it = e.begin();
                it != e.end(); it++) {
            std::cout << (*it) << " ";
        }
        std::cout << std::endl;
    } // --PBPmeas::show_all()

    ~PBPmeas() {
        if (std::verbose >= 5) std::cout << "In PBPmeas " << PMID << " Dtor" << std::endl;
    }
}; // --class PBPmeas

// unsigned PBPmeas::NEXT_PMID = 0;  // class variable initialization

/** a PBPmeaset is the set of point-by-point statistics for one execution of an algorithm */
class PBPmeaset {
    char algname[5]; // name of the algorithm (4 letters)  (not even necessary)
    char name[512]; // description of the algorithm
    execname_t execname;
    //  QString execname; // name of the execution:
    // snark can execute same algorithm many times with different parameters
    std::deque<PBPmeas> measurements; // deque is efficient for adding at end and iterating through
    std::vector<double> minima; // wrappers around simple built-in types: assume efficient
    std::vector<double> maxima;
    int minIter, maxIter;
    unsigned num_Es;
    static unsigned NEXT_PMSID;
    unsigned PMSID;
    bool completed;
    bool extrema_calculated;
public:

    PBPmeaset(execname_t line) {
        for (int i = 0; i < 4; i++) algname[i] = 'X';
        algname[4] = '\0'; // just in case
        PMSID = NEXT_PMSID++;
        num_Es = 0;
        minIter = maxIter = -1;
        completed = false;
        extrema_calculated = false;

        memset(name, 0, 512);
        char *p = line;
        char *q;
        q = &(name[0]);
        static const char *needle = "xecution name:";
        p = strstr(p, needle);
        if (p) {
            p += strlen(needle);
            while ((' ' == *p) || ('\t' == *p)) p++;
            for (int i = 0; (i < 490) && (*p) && ('\n' != *p) && ('\r' != *p); i++) {
                *q++ = *p++;
            }
        }
        *q = '\0'; // just to be sure
        /// std::cout << "exec name in PBPmeaset is: `" << name << "'" << std::endl;

        if (std::verbose >= 4) std::cout << "PBPmeaset " << PMSID << " ctor" << std::endl;
    } // --PBPmeaset ctor

    char* getAlgname() {
        return &(algname[0]);
    }
    // so this is a bit baroque... I guess we won't need the 4-letter algnames after all.

    execname_t getName() {
        return (&name[0]);
    }

    /** PBPmeaset::skip_algname() */
    measline_t skip_algname(measline_t line) {
        // my overkill implementation of strtok functionality (why? why?)
        measline_t cursor;
        //    std::cout<< "Input to skip_algname: " << std::endl << line << std::endl;;
        cursor = line;
        while ((' ' == (*cursor)) || ('\t' == (*cursor)))
            cursor++;
        if ((!(*cursor)) || !((*cursor) >= 'A') && ((*cursor) <= 'Z'))
            throw std::invalid_argument("bad line passed into PBPmeaset::skip_algname()");
        while ((*cursor) && ((*cursor) >= 'A') && ((*cursor) <= 'Z'))
            cursor++;
        //    std::cout<< "Returning from skip_algname: " << std::endl << cursor << std::endl;;
        return cursor;
    }

    /** PBPmeaset::set_algname() */
    measline_t set_algname(measline_t line) {
        measline_t cursor;
        cursor = line;
        while ((' ' == *cursor) || ('\t' == *cursor)) cursor++;
        if (!(*cursor >= 'A') && (*cursor <= 'Z'))
            throw std::invalid_argument("bad line passed into PBPmeaset::set_algname()");
        for (int i = 0; i < 4; i++)
            algname[i] = *cursor++;
        if ((algname[3] == '\t') || (algname[3] == ' ')) // alg is ART
            algname[3] = '\0';
        algname[4] = '\0'; // defensive
        return cursor;
    }

    //    void addMeasurement(PBPmeas& p) {
    //      if(completed)
    //        throw std::logic_error("Must NOT add to PBPmeaset once complete() is signalled");
    //      measurements.push_back(p);
    //    }

    /** PBPmeaset::addMeasurement() */
    void addMeasurement(measline_t line) {
        measline_t cursor = line;
        if (completed)
            throw std::logic_error("Must NOT add to PBPmeaset once complete() is signalled");
        if (measurements.empty()) // if this is the first line
            cursor = set_algname(cursor);
        else {
            //      std::cout<< "cursor:" << cursor << std::endl;
            cursor = skip_algname(cursor);
            //      std::cout<< "cursor:" << cursor << std::endl;
        }
        PBPmeas p(cursor);
        measurements.push_back(p);
    }

    void complete() {
        if (measurements.empty()) {
            throw std::logic_error("empty PBPmeaset refused to complete() itself");
        }
        // else
        if (completed)
            throw std::logic_error("complete() already signalled for this PBPmeaset, must NOT signal again.");
        // else
        completed = true;
    }

    /** PBPmeaset::isComplete() whether or not this point-by-point measurement set has been completed */
    bool isComplete() const {
        return completed;
    } // PBPmeaset::isComplete()

    /** ++PBPmeaset::get_num_Es() returns number of fields */
    unsigned get_num_Es() {
        if (measurements.empty()) {
            num_Es = 0;
        } else {
            num_Es = measurements[0].get_num_Es(); // guaranteed to be the same in all
        }
        return num_Es;
    } // --PBPmeaset::get_num_Es()

    /** PBPmeaset::calculate_extrema(), to be called as 0-arg or 3-arg
        returns true if input iteration range intersects with data's iteration range ("data domain")
        makes the assumption that iterations increase in a PBPmeaset (such as (*this) )
     */
    bool calculate_extrema(bool restrict_domain = false, int min_it_arg = 0, int max_it_arg = 0) {
        if (!completed)
            throw std::logic_error("Must NOT calculate extrema of PBPmeaset before complete() is signalled");
        // else
        if (0 == get_num_Es())
            throw std::logic_error("CANNOT calculate anything for PBPmeaset with no entries!");
        // else
        extrema_calculated = true; // intentionally set this to true even if return false
        std::deque<PBPmeas>::iterator it = measurements.begin();
        if (restrict_domain) {
            while ((it != measurements.end()) && (min_it_arg > (*it).getIter()))
                it++;
            if (it == measurements.end())
                return false;
        }
        minIter = maxIter = (*it).getIter();
        for (unsigned myInd = 0; myInd < num_Es; myInd++) {
            double cur_E = (*it).get_E(myInd);
            minima.push_back(cur_E);
            maxima.push_back(cur_E);
        }
        it++;
        for (; it != measurements.end(); it++) { // executes 0 or more times
            if (restrict_domain && (max_it_arg < (*it).getIter()))
                break;
            maxIter = (*it).getIter(); // last of the small ones
            for (unsigned myInd = 0; myInd < num_Es; myInd++) {
                double cur_E = (*it).get_E(myInd);
                if (cur_E < minima[myInd]) minima[myInd] = cur_E;
                if (cur_E > maxima[myInd]) maxima[myInd] = cur_E;
            }
        }
        return true;
    } // --PBPmeaset::calculate_extrema()

    int getMinIter() {
        if (!completed)
            throw std::logic_error("getMinIter() MUST not be called before complete() is signalled in PBPmeaset");
        if (extrema_calculated) // || calculate_extrema())
            return minIter; ///// control shouldn't reach the next line, just defensiveness
        else return -1;
    } // --PBPmeaset::getMinIter()

    int getMaxIter() {
        if (!completed)
            throw std::logic_error("getMaxIter() MUST not be called before complete() is signalled in PBPmeaset");
        if (extrema_calculated) // || calculate_extrema())
            return maxIter;
        else return -1;
    } // --PBPmeaset::getMaxIter()

    double getMinE(unsigned ind) {
        if (!completed)
            throw std::logic_error("getMinE() MUST not be called before complete() is signalled in PBPmeaset");
        // else
        if (!extrema_calculated)
            calculate_extrema(); // which calculates num_Es
        if (0 == num_Es)
            throw std::logic_error("num_Es cannot be 0 in PBPmeaset::getMinE()");
        // else
        if (ind >= num_Es)
            throw std::invalid_argument("bad parameter to getMinE");
        // else
        return minima[ind];
    } // --PBPmeaset::getMinE()

    double getMaxE(unsigned ind) {
        if (!completed)
            throw std::logic_error("getMaxE() MUST not be called before complete() is signalled in PBPmeaset");
        // else
        if (!extrema_calculated)
            calculate_extrema(); // which calculates num_Es
        if (0 == num_Es)
            throw std::logic_error("num_Es cannot be 0 in PBPmeaset::getMinE()");
        // else
        if (ind >= num_Es)
            throw std::invalid_argument("bad parameter to getMaxE");
        // else
        return maxima[ind];
    } // --PBPmeaset::getMaxE()

    /** shows the extrema of a PBP measurement set (due to a single execution of a single algorithm) */
    void show_extrema() {
        if (!completed)
            throw std::logic_error("Must NOT show extrema of PBPmeaset before complete() is signalled");
        if (!extrema_calculated) calculate_extrema(); // the order matters here
        if (minima.size() != maxima.size()) // won't really happen
            throw std::logic_error("minima and maxima of a PBPmeaset MUST have same size in show_extrema()");
        for (unsigned i = 0; i < minima.size(); i++) {
            std::cout << minima[i] << " " << maxima[i] << std::endl;
        }
    } // --PBPmeaset::show_extrema()

    /** shows all measurements of a measurement set (for debugging) */
    void show_all() {
        if (!completed)
            throw std::logic_error("Must NOT show_all for PBPmeaset before complete() is signalled");
        for (std::deque<PBPmeas>::iterator mit = measurements.begin();
                mit != measurements.end(); mit++) {
            mit->show_all();
        }
    } // --PBPmeaset::show_all()

    void addPBPVarLine(line_real_set_t& LRS, unsigned PBPYVar) {
        line_real_t LR(getName());
        for (std::deque<PBPmeas>::iterator it = measurements.begin();
                it != measurements.end(); it++) {
            LR.add(it->getIter(), it->get_E(PBPYVar));
        }
        if (!LR.empty()) LRS.add(LR);
    } // --PBPmeset::addPBPVarLine()

    /** PBPmeaset destructor */
    ~PBPmeaset() {
        if (std::verbose >= 4) std::cout << "PBPmeaset " << PMSID << " Dtor" << std::endl;
    }
}; // --class PBPmeaset

// unsigned PBPmeaset::NEXT_PMSID = 0; // class variable initialization

/** contains PBPmeasets and globmeasets that appear in a single EVALUATE result */
class eval_execution {
    std::deque<globmeaset> globmeasets;
    std::deque<PBPmeaset> PBPmeasets;
    bool completed;
    bool global_extrema_calculated;
    bool PBP_extrema_calculated;
    // extrama... none calculated until eval_execution is 'completed', all calculated at once.
    // first the global extrema
    int minIter, maxIter;
    double minAverage, maxAverage, minDistance, maxDistance, minRelerr, maxRelerr;
    double minVariance, maxVariance, minStddev, maxStddev, minResidual, maxResidual, minKullback, maxKullback;
    // then the PBP extrema...
    std::vector<double> PBP_minima; // minima and maxima of each E(0)..E(N-1), N not varying within an eval file.
    std::vector<double> PBP_maxima; // calculated over ALL algorithm execution/iterations in this EVALUATE
    unsigned num_Es; // 'N' in comment above
    phantomeas* phantomeasp; // will be added directly...
    // e.g., PBP_minima[2] is the minimum value of all E(2) in all PBPmeasets.
    static unsigned NEXT_EEID;
    unsigned EEID;
    char name[512];
public:

    eval_execution() {
        EEID = NEXT_EEID++;
        completed = global_extrema_calculated = PBP_extrema_calculated = false;
        phantomeasp = 0;
        num_Es = 0;
        minResidual = maxResidual = -1.0; // must initialize to negative value
        minKullback = maxKullback = -1.0;
        memset(name, 0, 512);
    } // --eval_execution ctor

    eval_execution(execname_t line) {
        ///    std::cout << "++eval_execution::eval_execution(), line is " << line;
        EEID = NEXT_EEID++;
        completed = global_extrema_calculated = PBP_extrema_calculated = false;
        phantomeasp = 0;
        num_Es = 0;
        minResidual = maxResidual = -1.0; // must initialize to negative value
        minKullback = maxKullback = -1.0; // must initialize to negative value
        char *p = line;
        char *q;
        q = &(name[0]);
        static const char *needle = "valuation name:";
        p = strstr(p, needle); // just in case of preceding whitespace
        memset(q, 0, 512); // sort of gratuitous
        q = &(name[0]);
        if (p) {
            p += strlen(needle);
            while ((' ' == *p) || ('\t' == *p)) p++;
            for (int i = 0; (i < 490) && (*p) && ('\n' != *p) && ('\r' != *p); i++) {
                *q++ = *p++;
            }
        }
        *q = '\0'; // just to be sure
        ///    std::cout << "--eval_execution::eval_execution(), name is " << (&(name[0])) << std::endl;
    } // --eval_execution ctor

    execname_t getName() {
        return static_cast<execname_t> (&(name[0]));
    } // --eval_execution::getName()

    void set_phantom(phantomeas* phanp) { // by pointer!!!
        phantomeasp = phanp;
    } // eval_execution::set_phantom()

    void new_globmeaset(execname_t line) {
        if (std::verbose >= 3)
            std::cout << "++ eval_execution::new_globmeaset()" << std::endl;
        if (completed)
            throw std::logic_error("eval_execution::complete() already signalled in eval_execution::new_globmeaset().");
        // else
        if (!globmeasets.empty()) {
            globmeasets.back().complete();
        }
        globmeaset g(line);
        globmeasets.push_back(g);
        if (std::verbose >= 3) std::cout << "-- eval_execution::new_globmeaset()" << std::endl;
    } // eval_execution::new_globmeaset()

    void add_to_globmeaset(measline_t line) {
        if (std::verbose >= 4)
            std::cout << "++ eval_execution::add_to_globmeaset(line)" << std::endl;
        if (completed)
            throw std::logic_error("eval_execution::complete() already signalled in eval_execution::add_to_globmeaset().");
        // else
        if (!hasGlobal())
            throw std::logic_error("attempting to add to globmeaset of eval_execution where none exist");
        globmeasets.back().addMeasurement(line);
        if (std::verbose >= 4)
            std::cout << "-- eval_execution::add_to_globmeaset(line)" << std::endl;
    }

    void new_PBPmeaset(execname_t line) {
        if (completed)
            throw std::logic_error("eval_execution::complete() already signalled in eval_execution::new_PBPmeaset().");
        // else
        if (!PBPmeasets.empty()) {
            PBPmeasets.back().complete();
        }
        PBPmeaset p(line);
        PBPmeasets.push_back(p);
    } // --eval_execution::new_PBPmeaset()

    void add_to_PBPmeaset(measline_t line) {
        if (completed)
            throw std::logic_error("eval_execution::complete() already signalled in eval_execution::add_to_PBPmeaset().");
        // else
        if (!hasPBP())
            throw std::logic_error("attempting to add to PBPmeaset of eval execution where none exist");
        // else
        PBPmeasets.back().addMeasurement(line); // add measurement to last PBPmeaset
    } // --eval_execution::add_to_PBPmeaset()

    unsigned get_num_globmeasets() const {
        return static_cast<unsigned> (globmeasets.size());
    }

    unsigned get_num_PBPmeasets() const {
        return static_cast<unsigned> (PBPmeasets.size());
    }

    bool hasGlobal() const {
        return (!(globmeasets.empty()));
    } // eval_execution::hasGlobal()

    bool hasPBP() const {
        return (!(PBPmeasets.empty()));
    } // eval_execution::hasPBP()

    /** Tell this eval_execution object that no more entries will be added to it, and calculations may begin */
    void complete() { // verb
        if ((!hasGlobal()) && (!hasPBP()))
            throw std::logic_error("eval_execution with NOTHING in it was asked to complete() itself!");
        // else
        if (completed)
            throw std::logic_error("attempt to complete() same eval execution twice!");
        // else
        completed = true;
        if (hasGlobal()) {
            if (std::verbose >= 3) std::cout << "calling calculate_global_extrema()" << std::endl;
            globmeasets.back().complete();
            calculate_global_extrema(); // a good place to do this. (is it?)
            if (std::verbose >= 3) std::cout << "returned from calculate_global_extrema()" << std::endl;
        }
        if (hasPBP()) {
            PBPmeasets.back().complete();
            if (std::verbose >= 3) std::cout << "calling calculate_PBP_extrema()" << std::endl;
            calculate_PBP_extrema(); // " " " " " "
            if (std::verbose >= 3) std::cout << "returned from calculate_PBP_extrema()" << std::endl;
        }
    } // --eval_execution::complete()

    /** eval_execution::isComplete()
        for debugging--ask the eval_execution object whether it has been told that no more data will be added to it */
    bool isComplete() const {
        return completed;
    } // --eval_execution::isComplete()

    /** eval_execution::calculate_global_extrema()
        calculates global extrema of ALL global measurement sets in this EVALUATE execution
        called with 0 or 3 args meaningfully: 0-arg means "calculate over ALL data"         */
    bool calculate_global_extrema(bool restrict_domain = false, int min_it_arg = 0, int max_it_arg = 0) {
        ///    if(global_extrema_calculated) return false;
        if (!completed)
            throw std::logic_error("Must NOT calculate extrema of globmeaset before complete() is signalled");
        // else
        if (!hasGlobal())
            throw std::logic_error("Must NOT calculate extrema of empty globmeaset");
        // else
        std::deque<globmeaset>::iterator it = globmeasets.begin();
        // guaranteed to exist and point to an actual globmeaset if control reached here
        // skip while current globmeaset's domain does not intersect with input domain
        // "allows" to recalculate global extrema regardless
        while ((it != globmeasets.end()) &&
                !((*it).calculate_extrema(restrict_domain, min_it_arg, max_it_arg))) {
            it++; // boolean expression of while has intentional and possibly necessary side effect
        }
        if (it == globmeasets.end())
            return false; // no globmeaset agrees with domain!  unbelievable!
        // else
        minIter = (*it).getMinIter(); // return value will automagically be right due to
        maxIter = (*it).getMaxIter(); // the above globmeaset::calculate_extrema() call (first that succeeded)

        minAverage = (*it).getMinAverage();
        minDistance = (*it).getMinDistance();
        minRelerr = (*it).getMinRelerr();
        minVariance = (*it).getMinVariance();
        minStddev = (*it).getMinStddev();

        maxAverage = (*it).getMaxAverage();
        maxDistance = (*it).getMaxDistance();
        maxRelerr = (*it).getMaxRelerr();
        maxVariance = (*it).getMaxVariance();
        maxStddev = (*it).getMaxStddev();

        double tmp;
        if (phantomeasp) {
            if (minAverage > (tmp = phantomeasp->getAverage())) minAverage = tmp;
            else if (maxAverage < tmp) maxAverage = tmp;

            if (minVariance > (tmp = phantomeasp->getVariance())) minVariance = tmp;
            else if (maxVariance < tmp) maxVariance = tmp;

            if (minStddev > (tmp = phantomeasp->getStddev())) minStddev = tmp;
            else if (maxStddev < tmp) maxStddev = tmp;
        }

        if ((*it).hasResidual()) { // not necessarily true in the first globmeaset
            double tmp;
            minResidual = (*it).getMinResidual(); // won't be nonnegative
            maxResidual = (*it).getMaxResidual(); // won't be nonnegative
            if (phantomeasp && (phantomeasp->hasResidual())) {
                if (minResidual > (tmp = phantomeasp->getResidual())) minResidual = tmp;
                else if (maxResidual < tmp) maxResidual = tmp;
            }
        }

        if ((*it).hasKullback()) { // not necessarily true in the first globmeaset
            double tmp;
            minKullback = (*it).getMinKullback(); // won't be nonnegative
            maxKullback = (*it).getMaxKullback(); // won't be nonnegative
            if (phantomeasp && (phantomeasp->hasKullback())) {
                if (minKullback > (tmp = phantomeasp->getKullback())) minKullback = tmp;
                else if (maxKullback < tmp) maxKullback = tmp;
            }
        }
        it++;
        for (; it != globmeasets.end(); it++) {
            if (!((*it).calculate_extrema(restrict_domain, min_it_arg, max_it_arg)))
                continue; // don't update as long as (*it)'s domain does not match the arguments
            int tmpi;
            if (minIter > (tmpi = ((*it).getMinIter()))) minIter = tmpi;
            if (maxIter < (tmpi = ((*it).getMaxIter()))) maxIter = tmpi;

            double tmp;
            if (minAverage > (tmp = ((*it).getMinAverage()))) minAverage = tmp;
            if (minDistance > (tmp = ((*it).getMinDistance()))) minDistance = tmp;
            if (minRelerr > (tmp = ((*it).getMinRelerr()))) minRelerr = tmp;
            if (minVariance > (tmp = ((*it).getMinVariance()))) minVariance = tmp;
            if (minStddev > (tmp = ((*it).getMinStddev()))) minStddev = tmp;

            if (maxAverage < (tmp = ((*it).getMaxAverage()))) maxAverage = tmp;
            if (maxDistance < (tmp = ((*it).getMaxDistance()))) maxDistance = tmp;
            if (maxRelerr < (tmp = ((*it).getMaxRelerr()))) maxRelerr = tmp;
            if (maxVariance < (tmp = ((*it).getMaxVariance()))) maxVariance = tmp;
            if (maxStddev < (tmp = ((*it).getMaxStddev()))) maxStddev = tmp;

            if ((*it).hasResidual()) {
                if (minResidual < 0) { // this is the first globmeaset with residual
                    minResidual = (*it).getMinResidual();
                    maxResidual = (*it).getMaxResidual();
                } else { // we have seen a globmeaset with residual before and now seeing another
                    double tmp;
                    if (minResidual > (tmp = ((*it).getMinResidual()))) minResidual = tmp;
                    if (maxResidual < (tmp = ((*it).getMaxResidual()))) maxResidual = tmp;
                }
            }
            if ((*it).hasKullback()) {
                if (minKullback < 0) { // this is the first globmeaset with kl distance
                    minKullback = (*it).getMinKullback();
                    maxKullback = (*it).getMaxKullback();
                } else { // we have seen a globmeaset with kl distance before and now seeing another
                    double tmp;
                    if (minKullback > (tmp = ((*it).getMinKullback()))) minKullback = tmp;
                    if (maxKullback < (tmp = ((*it).getMaxKullback()))) maxKullback = tmp;
                }
            }
        } // --foreach globmeaset
        global_extrema_calculated = true;
        return true;
    } // --eval_execution::calculate_global_extrema()

    /** These require global measurements' extrema to have been calculated */
    double getMinAverage() {
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMinAverage()");
        if (!completed)
            throw std::logic_error("attempt to get info from incomplete eval_execution");
        return minAverage;
    }

    double getMaxAverage() {
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMaxAverage()");
        if (!completed)
            throw std::logic_error("attempt to get info from incomplete eval_execution");
        return maxAverage;
    }

    double getMinDistance() {
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMinDistance()");
        if (!completed)
            throw std::logic_error("attempt to get info from incomplete eval_execution");
        return minDistance;
    } // --eval_execution::getMinDistance()

    double getMaxDistance() {
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMaxDistance()");
        if (!completed)
            throw std::logic_error("attempt to get info from incomplete eval_execution");
        return maxDistance;
    }

    double getMinRelerr() {
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMinRelerr()");
        if (!completed)
            throw std::logic_error("attempt to get info from incomplete eval_execution");
        return minRelerr;
    }

    double getMaxRelerr() {
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMaxRelerr()");
        if (!completed)
            throw std::logic_error("attempt to get info from incomplete eval_execution");
        return maxRelerr;
    } // --eval_execution::getMaxRelerr()

    double getMinVariance() {
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMinVariance()");
        if (!completed)
            throw std::logic_error("attempt to get info from incomplete eval_execution");
        return minVariance;
    }

    double getMaxVariance() {
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMaxVariance()");
        if (!completed)
            throw std::logic_error("attempt to get info from incomplete eval_execution");
        return maxVariance;
    }

    double getMinStddev() {
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMinStddev()");
        if (!completed)
            throw std::logic_error("attempt to get info from incomplete eval_execution");
        return minStddev;
    }

    double getMaxStddev() {
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMaxStddev()");
        if (!completed)
            throw std::logic_error("attempt to get info from incomplete eval_execution");
        return maxStddev;
    } // --eval_execution::getMaxStddev()

    bool hasResidual() { // true if some of its globmeasets have some residual entry
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMaxResidual()");
        if (!completed)
            throw std::logic_error("attempt to get info from incomplete eval_execution");
        return (-0.5 < minResidual);
    }

    bool hasKullback() { // true if some of its globmeasets have some kl distance entries
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMaxKullback()");
        if (!completed)
            throw std::logic_error("attempt to get info from incomplete eval_execution");
        return (-0.5 < minKullback);
    }

    // pre: do NOT call these before calling hasResidual() and making sure it's true

    double getMinResidual() {
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMinResidual()");
        if (!hasResidual())
            throw std::logic_error("hasResidual() is false in eval_execution::getMinResidual()");
        return minResidual;
    }

    double getMinKullback() {
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMinKullback()");
        if (!hasKullback())
            throw std::logic_error("hasKullback() is false in eval_execution::getMinKullback()");
        return minKullback;
    }

    double getMaxResidual() {
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMaxResidual()");
        if (!hasResidual())
            throw std::logic_error("hasResidual() is false in eval_execution::getMaxResidual()");
        return maxResidual;
    }

    double getMaxKullback() {
        if (!hasGlobal())
            throw std::logic_error("hasGlobal() is false in eval_execution::getMaxKullback()");
        if (!hasKullback())
            throw std::logic_error("hasKullback() is false in eval_execution::getMaxKullback()");
        return maxKullback;
    }

    /** eval_execution::calculate_PBP_extrema() */
    bool calculate_PBP_extrema(bool restrict_domain = false, int min_it_arg = 0, int max_it_arg = 0) {
        /// if(PBP_extrema_calculated) return; // not anymore.
        ///// else
        if (std::verbose >= 5) std::cout << "++ calculate_PBP_extrema()" << std::endl;
        //   PBP_minima.resize(0);
        //   PBP_maxima.resize(0);
        PBP_minima.clear();
        PBP_maxima.clear();
        if (!completed) {
            std::cerr << "Attempt to calculate_PBP_extrema() before signaling complete() to evaluate_execution";
            throw std::logic_error("Attempt to calculate_PBP_extrema() before signaling complete() to evaluate_execution");
        }
        // else
        if (!hasPBP()) {
            std::cerr << "Attempt to calculate_PBP_extrema() when no point-by-point data in this evaluate_execution";
            throw std::logic_error("Attempt to calculate_PBP_extrema() when no point-by-point data in this evaluate_execution");
        }
        // else
        PBP_extrema_calculated = true;
        std::deque<PBPmeaset>::iterator it = PBPmeasets.begin();
        num_Es = (*it).get_num_Es(); // this is the same throughout the file.  can use it even if we chuck the data.
        while ((it != PBPmeasets.end()) &&
                !((*it).calculate_extrema(restrict_domain, min_it_arg, max_it_arg))) {
            it++;
        }
        if (it == PBPmeasets.end())
            return false;
        // else
        minIter = (*it).getMinIter();
        maxIter = (*it).getMaxIter();
        PBP_minima.resize(num_Es);
        PBP_maxima.resize(num_Es);
        for (unsigned myInd = 0; myInd < num_Es; myInd++) {
            PBP_minima[myInd] = (*it).getMinE(myInd);
            //      PBP_maxima.push_back( (*it).getMaxE(myInd) );
            PBP_maxima[myInd] = (*it).getMaxE(myInd);
        }
        it++;
        for (; it != PBPmeasets.end(); it++) {
            // empty domain intersection
            if (!((*it).calculate_extrema(restrict_domain, min_it_arg, max_it_arg)))
                continue;
            //  now:  min_it_arg >= (*it).getMinIter() and max_it_arg <= (*it).getMaxIter()
            int tmpi;
            if (minIter > (tmpi = (*it).getMinIter())) minIter = tmpi;
            if (maxIter < (tmpi = (*it).getMaxIter())) maxIter = tmpi;
            for (unsigned myInd = 0; myInd < num_Es; myInd++) {
                double curMinE, curMaxE;
                if (PBP_minima[myInd] > (curMinE = (*it).getMinE(myInd))) PBP_minima[myInd] = curMinE;
                if (PBP_maxima[myInd] < (curMaxE = (*it).getMaxE(myInd))) PBP_maxima[myInd] = curMaxE;
            }
        }
        if (std::verbose >= 5) std::cout << "-- calculate_PBP_extrema()" << std::endl;
        return true;
    } // --eval_execution::calculate_PBP_extrema()

    /** for getMinIter() and getMaxIter(), global and PBP agree (given same restriction) */

    /** however, there is no guarantee that both are present.                            */
    int getMinIter() {
        if (!completed)
            throw std::logic_error("attempt to get info from incomplete eval_execution");
        // else
        if (!global_extrema_calculated && !PBP_extrema_calculated) {
            if (hasGlobal()) {
                calculate_global_extrema();
            } else if (hasPBP()) {
                calculate_PBP_extrema();
            } else throw std::logic_error("no Global, no PBP in eval_execution::getMinIter()");
        }
        return minIter;
    }

    int getMaxIter() {
        if (!completed)
            throw std::logic_error("attempt to get info from incomplete eval_execution");
        // else
        if (!global_extrema_calculated && !PBP_extrema_calculated) {
            if (hasGlobal()) {
                calculate_global_extrema();
            } else if (hasPBP()) {
                calculate_PBP_extrema();
            } else throw std::logic_error("no Global, no PBP in eval_execution::getMaxIter()");
        }
        return maxIter;
    }

    /** eval_execution::get_num_Es() with possible nasty side effect of recalculating PBP extrema! */
    unsigned get_num_Es() {
        if (!completed)
            throw std::logic_error("get_num_Es() MUST not be called before complete() is signalled in eval_execution");
        if (!PBP_extrema_calculated)
            calculate_PBP_extrema();
        return num_Es; // due to the "side effect" of calling calculate_PBP_extrema
    }

    /** eval_execution::getMinE() to get minimum of a PBP measure attribute through EVALUATE */
    double getMinE(unsigned ind) {
        if (!completed)
            throw std::logic_error("getMinE() MUST not be called before complete() is signalled in eval_execution");
        // else
        if (!PBP_extrema_calculated)
            calculate_PBP_extrema(); // which calculates num_Es
        if (0 == num_Es)
            throw std::logic_error("num_Es cannot be 0 in PBPmeaset::getMinE()");
        // else
        if (ind >= num_Es)
            throw std::invalid_argument("bad parameter to getMinE");
        // else
        return PBP_minima[ind];
    } // eval_execution::getMinE()

    double getMaxE(unsigned ind) {
        if (!completed)
            throw std::logic_error("getMinE() MUST not be called before complete() is signalled in PBPmeaset");
        // else
        if (!PBP_extrema_calculated)
            calculate_PBP_extrema(); // which calculates num_Es
        if (0 == num_Es)
            throw std::logic_error("num_Es cannot be 0 in PBPmeaset::getMaxE()");
        // else
        if (ind >= num_Es)
            throw std::invalid_argument("bad parameter to getMaxE");
        // else
        return PBP_maxima[ind];
    } // eval_execution::getMaxE()

    /** debug routine: shows all values in the global measurement sets "recursively" */
    void show_all_globs() {
        if (!completed) {
            //        std::cout << "(cout) about to bail" << std::endl;
            //        std::cerr << "(cerr) about to bail" << std::endl;
            throw std::logic_error("MUST not call show_all_globs() before signalling complete() to an eval_execution");
        }
        // else
        if (globmeasets.empty()) {
            std::cout << "## No glob sets to show in eval exec " << EEID << std::endl;
        } else {
            std::cout << "=>>=showing all globmeasets in eval exec " << EEID << std::endl;
            for (unsigned i = 0; i < get_num_globmeasets(); i++) {
                std::cout << "Showing globmeaset " << i << ":" << std::endl;
                globmeasets[i].show_all();
            }
            std::cout << "=<<=showed all globmeasets in eval exec " << EEID << std::endl;
        }
    } // --eval_execution::show_all_globs()

    /** debug routine: shows all values in the global measurement sets "recursively" */
    void show_all_PBPs() {
        if (!completed)
            throw std::logic_error("MUST not call show_all_ppbs() before signalling complete() to eval_execution");
        // else
        if (PBPmeasets.empty()) {
            std::cout << "## No PBP sets to show in eval exec " << EEID << std::endl;
        } else {
            std::cout << "=>>=showing all PBPmeasets in eval exec " << EEID << std::endl;
            for (unsigned i = 0; i < get_num_PBPmeasets(); i++) {
                std::cout << "Showing PBPmeaset " << i << ": (algname ["
                        << PBPmeasets[i].getAlgname() << "])" << std::endl;
                PBPmeasets[i].show_all();
            }
            std::cout << "=<<=showed all PBPmeasets in eval exec " << EEID << std::endl;
        }
    } // --eval_execution::show_all_PBPs()

    /** eval_execution::show_extrema() */
    void show_extrema() {
        if (!completed)
            throw std::logic_error("Must NOT show extrema of eval_execution before complete() is signalled");
        std::cout << "Showing extrema in eval_execution_set among selected eval_executions." << std::endl;
        bool successful = false;
        if (hasGlobal() && calculate_global_extrema()) {
            std::cout << "after global calculation: minIter, maxIter: " << minIter << " " << maxIter << std::endl;
            std::cout << "minAverage, maxAverage, minDistance, maxDistance, minRelerr, maxRelerr:" << std::endl
                    << minAverage << " " << maxAverage << " " << minDistance << " " << maxDistance << " " << minRelerr << " " << maxRelerr;
            std::cout << std::endl;
            std::cout << "minVariance, maxVariance, minStddev, maxStddev, minResidual, maxResidual, minKullback, maxKullback:" << std::endl
                    << minVariance << " " << maxVariance << " " << minStddev << " " << maxStddev << " " << minResidual << " " << maxResidual << " " << minKullback << " " << maxKullback;
            std::cout << std::endl;
            successful = true;
        }

        if (hasPBP() && calculate_PBP_extrema()) {
            std::cout << "after PBP calculation: minIter, maxIter: " << minIter << " " << maxIter << std::endl;
            for (unsigned i = 0; i < num_Es; i++) {
                std::cout << "PBP_minima[" << i << "] = " << PBP_minima[i] << "; "
                        << "PBP_maxima[" << i << "] = " << PBP_maxima[i] << std::endl;
            }
            successful = true;
        }
        if (!successful) {
            std::cout << "Attempts to calculate both global and PBP extrema were unsuccessful in eval_execution." << std::endl;
        }
    } // --eval_execution::show_extrema()

    void show_all() {
        std::cout << "==>>>==entering show_all() for eval exec " << EEID << std::endl;
        show_all_globs();
        show_all_PBPs();
        std::cout << "==<<<==leaving show_all() for eval exec " << EEID << std::endl;
    } // --eval_execution::show_all()

    void addPBPVarLines(line_real_set_t& LRS, unsigned PBPYVar) {
        for (std::deque<PBPmeaset>::iterator it = PBPmeasets.begin();
                it != PBPmeasets.end(); it++) {
            it->addPBPVarLine(LRS, PBPYVar);
        }
    } // --eval_execution::addPBPVarLines()

    void addGlobVarLines(line_real_set_t& LRS, globalYVar_t globalYVar) {
        for (std::deque<globmeaset>::iterator it = globmeasets.begin();
                it != globmeasets.end(); it++) {
            it->addGlobVarLine(LRS, globalYVar);
        }
    } // --eval_execution::addGlobVarLines()

}; // --class eval_execution

// unsigned eval_execution::NEXT_EEID = 0; // class variable initialization

/** eval_execution_set: superstructure capable of holding information 
    given by many eval executions, an entire eval08 file                 */
//  class eval_execution_set {
//    std::deque<eval_execution> execs; // entire output of snark14 EVALUATE executions
//    std::vector<bool> selected; // holds which execs are selected by snark14Display user
//    phantomeas* phantomeasp;
//    bool completed; // holds whether creation of execs is complete (is this really necessary here?)
//    bool extrema_calculated; // holds whether extrema are calculated (ditto.)
//    bool same_x_forall;  // for iter, plotted on x-axis
//    bool same_y_forall;  // for whatever it is that is requested for plotting on y-axis
//    // (almost the same as for eval_execution) extrama...
//    // first the global extrema
//    int minIter, maxIter;
//    double minAverage, maxAverage, minDistance, maxDistance, minRelerr, maxRelerr;
//    double minVariance, maxVariance, minStddev, maxStddev, minResidual, maxResidual;
//    // then the PBP extrema...
//    std::vector<double> PBP_minima; // minima and maxima of each E(0)..E(N-1), N not varying within an eval file.
//    std::vector<double> PBP_maxima; // to be calculated over ALL EVALUATE calls, depending.
//    unsigned num_Es; // 'N' in comment above
//    // e.g., PBP_minima[2] is the minimum value of all E(2) in all PBPmeasets in all execs.

eval_execution_set::eval_execution_set() {
    completed = extrema_calculated = false;
    same_x_forall = same_y_forall = false; // initialize to use "locally" calculated domain/range bounds
    selectionHasResidualvar = false;
    selectionHasKullbackvar = false;
    phantomeasp = 0;
} // eval_execution_set() ctor

//   void new_exec(eval_execution &ev) {
//     // pseudocode
//     execs.push_back(ev);
//     selected.push_back(false);
//   }

// void eval_execution_set::new_exec() {
//   if (!execs.empty()) {
//     execs.back().set_phantom(phantomeasp);
//     execs.back().complete();
//   }
//   eval_execution exec;
//   execs.push_back(exec);
//   selected.push_back(false);
// }

void eval_execution_set::new_exec(execname_t name) {
    if (!execs.empty()) {
        execs.back().set_phantom(phantomeasp);
        execs.back().complete();
    }
    eval_execution exec(name);
    execs.push_back(exec);
    selected.push_back(false);
}

void eval_execution_set::new_globmeaset(execname_t line) {
    if (execs.empty())
        throw std::logic_error("call to eval_execution_set::new_globmeaset() before ::new_exec()");
    // else
    execs.back().new_globmeaset(line);
}

void eval_execution_set::add_to_globmeaset(measline_t line) {
    if (execs.empty())
        throw std::logic_error("call to eval_execution_set::add_to_globmeaset() before ::new_exec()");
    execs.back().add_to_globmeaset(line);
}

void eval_execution_set::new_PBPmeaset(execname_t line) {
    if (execs.empty())
        throw std::logic_error("call to eval_execution_set::new_PBPmeaset() before ::new_exec()");
    execs.back().new_PBPmeaset(line);
}

void eval_execution_set::add_to_PBPmeaset(measline_t line) {
    if (execs.empty())
        throw std::logic_error("call to eval_execution_set::add_to_PBPmeaset() before ::new_exec()");
    execs.back().add_to_PBPmeaset(line);
}

/** eval_execution_set::set_phantomeas() sets _the_ phantom measurement of the eval_execution_set
    pre: all phantom-related lines in the eval file are identical
    barring the possibility of there being residual in some but not others
    this function sort of "updates" the value of phantom.
    logically, to be considered part of every eval_execution that has at least one globmeaset:
    (for applicable fields) both for the purpose of plotting and calculating extrema.
    if exists (!=0), DO NOT CHANGE phantomeasp OR YOU'LL SEGFAULT!
 */
void eval_execution_set::set_phantomeas(measline_t line) {

    if (phantomeasp && phantomeasp->hasResidual() && phantomeasp->hasKullback()) return;
    // else
    if (!phantomeasp) {
        phantomeasp = new phantomeas(line);
        return;
    }
    // else
    phantomeas locphan(line);
    if (locphan.hasResidual()) {
        phantomeasp->setNewResidual(locphan.getResidual()); // set by value
    }
    if (locphan.hasKullback()) {
        phantomeasp->setNewKullback(locphan.getKullback()); // set by value
    }
} // --eval_execution_set::set_phantomeas()

bool eval_execution_set::hasPhantom() {
    return (0 != phantomeasp);
}

bool eval_execution_set::calculate_global_extrema() {
    return calculate_global_extrema(false, 0, 0);
}

/** eval_execution_set::calculate_global_extrema()  */
bool eval_execution_set::calculate_global_extrema(bool restrict_domain, int min_it_arg, int max_it_arg) {
    if (!completed)
        throw std::logic_error("eval_execution_set::calculate_extrema() cannot be called on incomplete set");
    if (selected.size() != execs.size())
        throw std::logic_error("selected does not match execs in size");
    minResidual = maxResidual = -1.0;
    minKullback = maxKullback = -1.0;
    bool first_one = true;
    unsigned j = 0;
    for (std::deque<eval_execution>::iterator it = execs.begin();
            it != execs.end(); j++, it++) {
        if (selected[j] && ((*it).hasGlobal()) && ((*it).calculate_global_extrema(restrict_domain, min_it_arg, max_it_arg))) {
            /// std::cout << "eval_execution_set::calculate_global_extrema() j = " << j << std::endl;
            if (first_one) { // for first selected eval_exec with global
                // it is the first one and non-empty intersects with argument domain, if restricted
                minIter = (*it).getMinIter();
                maxIter = (*it).getMaxIter();
                /// std::cout << "minIter: " <<  minIter << ", maxIter: " << maxIter << std::endl;
                minAverage = (*it).getMinAverage();
                minDistance = (*it).getMinDistance();
                minRelerr = (*it).getMinRelerr();
                minVariance = (*it).getMinVariance();
                minStddev = (*it).getMinStddev();

                maxAverage = (*it).getMaxAverage();
                maxDistance = (*it).getMaxDistance();
                maxRelerr = (*it).getMaxRelerr();
                maxVariance = (*it).getMaxVariance();
                maxStddev = (*it).getMaxStddev();

                if ((*it).hasResidual()) { // not necessarily true in the first evaluate
                    selectionHasResidualvar = true;
                    minResidual = (*it).getMinResidual();
                    maxResidual = (*it).getMaxResidual();
                }
                if ((*it).hasKullback()) { // not necessarily true in the first evaluate
                    selectionHasKullbackvar = true;
                    minKullback = (*it).getMinKullback();
                    maxKullback = (*it).getMaxKullback();
                }
                first_one = false;
            } else { // for all other selected eval_execs with global with matching domain
                int tmpi;
                if (minIter > (tmpi = (*it).getMinIter())) minIter = tmpi;
                if (maxIter < (tmpi = (*it).getMaxIter())) maxIter = tmpi;
                double tmp;
                if (minAverage > (tmp = ((*it).getMinAverage()))) minAverage = tmp;
                if (minDistance > (tmp = ((*it).getMinDistance()))) minDistance = tmp;
                if (minRelerr > (tmp = ((*it).getMinRelerr()))) minRelerr = tmp;
                if (minVariance > (tmp = ((*it).getMinVariance()))) minVariance = tmp;
                if (minStddev > (tmp = ((*it).getMinStddev()))) minStddev = tmp;

                if (maxAverage < (tmp = ((*it).getMaxAverage()))) maxAverage = tmp;
                if (maxDistance < (tmp = ((*it).getMaxDistance()))) maxDistance = tmp;
                if (maxRelerr < (tmp = ((*it).getMaxRelerr()))) maxRelerr = tmp;
                if (maxVariance < (tmp = ((*it).getMaxVariance()))) maxVariance = tmp;
                if (maxStddev < (tmp = ((*it).getMaxStddev()))) maxStddev = tmp;

                if ((*it).hasResidual()) {
                    if (minResidual < 0) { // this is the first globmeaset with residual
                        minResidual = (*it).getMinResidual();
                        maxResidual = (*it).getMaxResidual();
                    } else { // we have seen a globmeaset with residual before and now seeing another
                        double tmp;
                        if (minResidual > (tmp = ((*it).getMinResidual()))) minResidual = tmp;
                        if (maxResidual < (tmp = ((*it).getMaxResidual()))) maxResidual = tmp;
                    }
                }
                if ((*it).hasKullback()) {
                    if (minKullback < 0) { // this is the first globmeaset with kl distance
                        minKullback = (*it).getMinKullback();
                        maxKullback = (*it).getMaxKullback();
                    } else { // we have seen a globmeaset with kl distance before and now seeing another
                        double tmp;
                        if (minKullback > (tmp = ((*it).getMinKullback()))) minKullback = tmp;
                        if (maxKullback < (tmp = ((*it).getMaxKullback()))) maxKullback = tmp;
                    }
                }
            } // --not the first one
        } // --selected evaluate with global
    } // --foreach EVALUATE execution
    return (!first_one);
    // return true if some execution has global measurements whose domain matches input domain restriction
} // --eval_exec_set::calculate_global_extrema()

bool eval_execution_set::calculate_PBP_extrema() {
    return calculate_PBP_extrema(false, 0, 0);
}

/** eval_execution_set::calculate_PBP_extrema()  */
bool eval_execution_set::calculate_PBP_extrema(bool restrict_domain, int min_it_arg, int max_it_arg) {
    if (!completed)
        throw std::logic_error("eval_execution_set::calculate_extrema() cannot be called on incomplete set");
    if (selected.size() != execs.size())
        throw std::logic_error("selected does not match execs in size");
    num_Es = 0;
    //    PBP_minima.resize(0);
    //    PBP_maxima.resize(0);
    PBP_minima.clear();
    PBP_maxima.clear();
    bool first_one = true;
    unsigned j = 0;
    for (std::deque<eval_execution>::iterator it = execs.begin();
            it != execs.end(); j++, it++) {
        if ((selected[j]) && ((*it).hasPBP()) && ((*it).calculate_PBP_extrema(restrict_domain, min_it_arg, max_it_arg))) {
            if (first_one) {
                minIter = (*it).getMinIter();
                maxIter = (*it).getMaxIter();
                num_Es = (*it).get_num_Es(); // this is the same throughout the file.
                PBP_minima.resize(num_Es);
                PBP_maxima.resize(num_Es);
                for (unsigned myInd = 0; myInd < num_Es; myInd++) {
                    PBP_minima[myInd] = (*it).getMinE(myInd);
                    PBP_maxima[myInd] = (*it).getMaxE(myInd);
                }
                first_one = false;
            } else {
                int tmpi;
                if (minIter > (tmpi = (*it).getMinIter())) minIter = tmpi;
                if (maxIter < (tmpi = (*it).getMaxIter())) maxIter = tmpi;
                for (unsigned myInd = 0; myInd < num_Es; myInd++) {
                    double tmp;
                    if (PBP_minima[myInd] > (tmp = (*it).getMinE(myInd))) PBP_minima[myInd] = tmp;
                    if (PBP_maxima[myInd] < (tmp = (*it).getMaxE(myInd))) PBP_maxima[myInd] = tmp;
                }
            } // not first one
        } // --selected evaluate with point
    } // --for
    return (!first_one);
    // return true if any selected execution has PBP measurements and input domain restriction matches their domain
} // eval_execution_set::calculate_PBP_extrema()

execname_t eval_execution_set::getExecName(unsigned idx) {
    return execs[idx].getName();
}

/** by "for all", the selected set of EVALUATE executions are meant.
    setSameXForAll(true) to use "global" domain for each plot */
//void eval_execution_set::setSameXForAll(bool val = true) {

void eval_execution_set::setSameXForAll(bool val) {
    same_x_forall = val;
}

//void eval_execution_set::setSameYForAll(bool val = true) {

void eval_execution_set::setSameYForAll(bool val) {
    same_y_forall = val;
}

bool eval_execution_set::getSameXForAll() {
    return same_x_forall;
}

bool eval_execution_set::getSameYForAll() {
    return same_y_forall;
}

// i don't even know if this idea will make it to the production version
//void eval_execution_set::useSameExtremaForAll(bool val = true) {

void eval_execution_set::useSameExtremaForAll(bool val) {
    setSameXForAll(val);
    setSameYForAll(val);
}

void eval_execution_set::useDifferentExtremaForEach() {
    useSameExtremaForAll(false);
}

void eval_execution_set::complete() {
    if (execs.empty())
        throw std::logic_error("attempt to compete() empty eval_execution_set!");
    // else
    execs.back().set_phantom(phantomeasp);
    execs.back().complete();
    completed = true;
}

unsigned eval_execution_set::num_eval_execs() {
    return static_cast<unsigned> (execs.size());
}

void eval_execution_set::deselect_exec(unsigned idx) {
    if (!completed)
        throw std::logic_error("Cannot deselect_exec() without having a completed eval_execution_set.");
    if (idx >= num_eval_execs())
        throw std::invalid_argument("exec index out of range in eval_execution_set::deselect_exec()");
    selected[idx] = false;
    extrema_calculated = false; // previous extrema calcs annulled
}

void eval_execution_set::select_exec(unsigned idx) { // everything is deselected by default
    if (!completed)
        throw std::logic_error("Cannot select_exec() without having a completed eval_execution_set.");
    if (idx >= num_eval_execs())
        throw std::invalid_argument("exec index out of range in eval_execution_set::deselect_exec()");
    selected[idx] = true;
    extrema_calculated = false; // previous extrema calcs annulled
} // eval_execution_set::select_exec()

void eval_execution_set::deselect_all_execs() {
    if (!completed)
        throw std::logic_error("Cannot deselect_all_execs() without having a completed eval_execution_set.");
    for (std::vector<bool>::iterator it = selected.begin(); it != selected.end(); it++) {
        *it = false;
    }
    extrema_calculated = false; // previous extrema calcs annulled
}

void eval_execution_set::select_all_execs() {
    if (!completed)
        throw std::logic_error("Cannot select_all_execs() without having a completed eval_execution_set.");
    for (std::vector<bool>::iterator it = selected.begin(); it != selected.end(); it++) {
        *it = true;
    }
    extrema_calculated = false; // previous extrema calcs annulled
}

bool eval_execution_set::selectionHasGlobal() {
    for (unsigned idx = 0; idx < num_eval_execs(); idx++) {
        if ((selected[idx]) && (execs[idx].hasGlobal())) {
            return true;
        }
    }
    return false;
}

bool eval_execution_set::selectionHasResidual() {
    for (unsigned idx = 0; idx < num_eval_execs(); idx++) {
        if ((selected[idx]) && (execs[idx].hasGlobal()) && (execs[idx].hasResidual())) {
            return true;
        }
    }
    return false;
}

bool eval_execution_set::selectionHasKullback() {
    for (unsigned idx = 0; idx < num_eval_execs(); idx++) {
        if ((selected[idx]) && (execs[idx].hasGlobal()) && (execs[idx].hasKullback())) {
            return true;
        }
    }
    return false;
}

bool eval_execution_set::selectionHasPBP() {
    for (unsigned idx = 0; idx < num_eval_execs(); idx++) {
        if ((selected[idx]) && (execs[idx].hasPBP())) {
            return true;
        }
    }
    return false;
}

/** eval_execution_set::get_num_Es() returns:
    number of PBP fields IF execution selection includes a PBP
                       0 otherwise (should? throw exception)
 */
unsigned eval_execution_set::get_num_Es() {
    for (unsigned idx = 0; idx < num_eval_execs(); idx++) {
        if ((selected[idx]) && (execs[idx].hasPBP())) {
            return execs[idx].get_num_Es();
        }
    }
    return 0;
}

// really throw exception... arkh.

int eval_execution_set::getMinIterViaGlobal() {
    if (calculate_global_extrema())
        return minIter;
    return -1;
}

int eval_execution_set::getMaxIterViaGlobal() {
    if (calculate_global_extrema())
        return maxIter;
    return -1;
}

int eval_execution_set::getMinIterViaPBP() {
    if (calculate_PBP_extrema())
        return minIter;
    return -1;
}

int eval_execution_set::getMaxIterViaPBP() {
    if (calculate_PBP_extrema())
        return maxIter;
    return -1;
}

double eval_execution_set::getMinYGlobal(globalYVar_t globalYVar) {
    if (!calculate_global_extrema()) return -1; // throw exception!
    switch (globalYVar) {
        case AVERAGE:
            return minAverage;
            break;
        case DISTANCE:
            return minDistance;
            break;
        case RELERR:
            return minRelerr;
            break;
        case VARIANCE:
            return minVariance;
            break;
        case STDDEV:
            return minStddev;
            break;
        case RESIDUAL:
            if (selectionHasResidual())
                return minResidual;
        case KULLBACK:
            if (selectionHasKullback())
                return minKullback;
            break;
    } // switch
    // throw exception!
    return -1;
}

double eval_execution_set::getMaxYGlobal(globalYVar_t globalYVar) {
    if (!calculate_global_extrema()) return -1; // throw exception!
    switch (globalYVar) {
        case AVERAGE:
            return maxAverage;
            break;
        case DISTANCE:
            return maxDistance;
            break;
        case RELERR:
            return maxRelerr;
            break;
        case VARIANCE:
            return maxVariance;
            break;
        case STDDEV:
            return maxStddev;
            break;
        case RESIDUAL:
            if (selectionHasResidual())
                return maxResidual;
        case KULLBACK:
            if (selectionHasKullback())
                return maxKullback;
            break;
    } // --switch
    // throw exception!
    return -1;
} // --eval_execution_set::getMaxYGlobal(globalYVar_t globalYVar)

double eval_execution_set::getMinYPBP(unsigned idx) {
    //   if(idx >= num_Es) {
    //     throw std::logic_error("blah");
    //   }
    calculate_PBP_extrema();
    return PBP_minima[idx];
}

double eval_execution_set::getMaxYPBP(unsigned idx) {
    calculate_PBP_extrema();
    return PBP_maxima[idx];
}

/** eval_execution_set::show_all() and show_selected() are for debugging */
void eval_execution_set::show_all() {
    if (hasPhantom()) {
        phantomeasp->show();
    }
    for (std::deque<eval_execution>::iterator it = execs.begin(); it != execs.end(); it++) {
        (*it).show_all();
    } // --foreach eval_execution
} // eval_execution_set::show_all()

void eval_execution_set::show_selected() {
    unsigned j = 0;
    for (std::deque<eval_execution>::iterator it = execs.begin();
            it != execs.end(); j++, it++) {
        if (selected[j]) {
            (*it).show_all();
        }
    } // --foreach eval_execution

} // eval_execution_set::show_selected()

/** eval_execution_set::show_extrema() [mainly for debugging]
 *  huge side effect: recalculates all extrema without domain restriction */
void eval_execution_set::show_extrema() {
    if (!completed)
        throw std::logic_error("Must NOT show extrema of eval_execution_set before complete() is signalled");
    std::cout << "Showing extrema in eval_execution_set among selected eval_executions." << std::endl;

    // hack...
    if (hasPhantom()) {
        phantomeasp->show();
    }

    bool successful = false;
    if (calculate_global_extrema()) {
        std::cout << "after global calculation: minIter, maxIter: " << minIter << " " << maxIter << std::endl;
        std::cout << "minAverage, maxAverage, minDistance, maxDistance, minRelerr, maxRelerr:" << std::endl
                << minAverage << " " << maxAverage << " " << minDistance << " " << maxDistance << " " << minRelerr << " " << maxRelerr;
        std::cout << std::endl;
        std::cout << "minVariance, maxVariance, minStddev, maxStddev, minResidual, maxResidual, minKullback, maxKullback:" << std::endl
                << minVariance << " " << maxVariance << " " << minStddev << " " << maxStddev << " " << minResidual << " " << maxResidual << " " << minKullback << " " << maxKullback;
        std::cout << std::endl;
        successful = true;
    }

    if (calculate_PBP_extrema()) {
        std::cout << "after PBP calculation: minIter, maxIter: " << minIter << " " << maxIter << std::endl;
        for (unsigned i = 0; i < num_Es; i++) {
            std::cout << "PBP_minima[" << i << "] = " << PBP_minima[i] << "; "
                    << "PBP_maxima[" << i << "] = " << PBP_maxima[i] << std::endl;
        }
        successful = true;
    }

    if (!successful) {
        std::cout << "Attempts to calculate both global and PBP extrema were unsuccessful." << std::endl;
    }
} // --eval_execution_set::show_extrema()

void eval_execution_set::addPBPVarLines(line_real_set_t& LRS, unsigned PBPYVar, unsigned ee_index) {
    execs[ee_index].addPBPVarLines(LRS, PBPYVar); // this is 'inefficient' addressing, so what?
}

void eval_execution_set::addGlobVarLines(line_real_set_t& LRS, globalYVar_t globalYVar, unsigned ee_index) {
    double phanval;
    // phantom first
    if (phantomeasp) {
        phanval = phantomeasp->get(globalYVar);
    } else {
        phanval = -2.0;
    }
    if (phanval > -0.5) {
        line_real_t phantomline("phantom", LRS.getMinX(), LRS.getMaxX(), phanval);
        LRS.add(phantomline);
    }
    // then the other lines
    execs[ee_index].addGlobVarLines(LRS, globalYVar);
} // --eval_execution_set::addGlobVarLines()

void eval_execution_set::spawnDialogsGlob(QWidget* parent, globalYVar_t globalYVar) {
    if (std::verbose >= 2) {
        std::cout << "in eval_execution_set::spawnDialogsGlob()" << std::endl;
    }
    int posx = parent->pos().x();
    int posy = parent->pos().y();
    unsigned ee_index = 0;
    for (std::deque<eval_execution>::iterator it = execs.begin();
            it != execs.end(); ee_index++, it++) {
        // if selected and has global and if selected var is residual then has residual
        if ((selected[ee_index]) && ((*it).hasGlobal()) && ((globalYVar != RESIDUAL) || ((*it).hasResidual())) && ((globalYVar != KULLBACK) || ((*it).hasKullback()))) {
            posx += 20;
            posy += 20;
            ///      it->show_extrema(); // might have side-effects
            it->calculate_global_extrema();
            int minX = it->getMinIter(); // to be the defaults in the range dialog window
            int maxX = it->getMaxIter();
            double minY = 0, maxY = 0; // initialize to shut up the compiler
            switch (globalYVar) {
                case AVERAGE: minY = it->getMinAverage();
                    maxY = it->getMaxAverage();
                    break;
                case DISTANCE: minY = it->getMinDistance();
                    maxY = it->getMaxDistance();
                    break;
                case RELERR: minY = it->getMinRelerr();
                    maxY = it->getMaxRelerr();
                    break;
                case VARIANCE: minY = it->getMinVariance();
                    maxY = it->getMaxVariance();
                    break;
                case STDDEV: minY = it->getMinStddev();
                    maxY = it->getMaxStddev();
                    break;
                case RESIDUAL: minY = it->getMinResidual();
                    maxY = it->getMaxResidual();
                    break;
                case KULLBACK: minY = it->getMinKullback();
                    maxY = it->getMaxKullback();
                    break;
                    // default: never happens
            }
            if (std::verbose >= 2) {
                std::cout << "in eval_execution_set::spawnDialogsGlob() minY is "
                        << minY << " and maxY is " << maxY << std::endl;
            }
            const char* ccp = it->getName();
            chooseRangesDialog* crdp =
                    new chooseRangesDialog(parent, ccp, false, 0, posx, posy,
                    ee_index, false, 0, globalYVar,
                    minX, maxX, minY, maxY);
            crdp->show();
        } // if
    } // for
} // --eval_execution_set::spawnDialogsGlob()

/** spawn dialog box to get range for each selected eval execution with PBP */
void eval_execution_set::spawnDialogsPBP(QWidget* parent, unsigned PBPYVar) {
    if (std::verbose >= 2) {
        std::cout << "in eval_execution_set::spawnDialogsPBP()" << std::endl;
    }
    int posx = parent->pos().x();
    int posy = parent->pos().y();
    unsigned ee_index = 0;
    for (std::deque<eval_execution>::iterator it = execs.begin();
            it != execs.end(); ee_index++, it++) {
        if ((selected[ee_index]) && ((*it).hasPBP())) {
            posx += 20;
            posy += 20;
            it->calculate_PBP_extrema();
            int minX = it->getMinIter(); // to be the defaults in the range dialog window
            int maxX = it->getMaxIter();
            double minY = it->getMinE(PBPYVar);
            double maxY = it->getMaxE(PBPYVar);
            const char* ccp = it->getName();
            chooseRangesDialog* crdp =
                    new chooseRangesDialog(parent, ccp, false, 0, posx, posy,
                    ee_index, true, PBPYVar, AVERAGE,
                    minX, maxX, minY, maxY);
            crdp->show();
        } // if
    } // --for
} // --eval_execution_set::spawnDialogsPBP()

void eval_execution_set::spawnAllPlotsGlob(QWidget* parent,
        globalYVar_t globalYVar, int userMinXVal, int userMaxXVal,
        double userMinYVal, double userMaxYVal, bool isGrayScale) {
    /// std::cout << "in eval_execution_set::spawnAllPlotsGlob()" << std::endl;
    //  int PMWidth  = 650; // pixmap width, etc
    //  sd_line_t::WIDTH  = 650;
    int PMWidth = sd_line_t::WIDTH;
    //  int PMHeight = 550;
    //  sd_line_t::HEIGHT = 550;
    int PMHeight = sd_line_t::HEIGHT;
    int PMDepth = 24; // number of colors?
    unsigned ee_index = 0;
    int x = parent->pos().x();
    int y = parent->pos().y();
    // block for doing phan
    double phanval;
    if (phantomeasp) {
        phanval = phantomeasp->get(globalYVar);
    } else {
        phanval = -2.0;
    }
    QString YVarName;
    switch (globalYVar) {
        case AVERAGE: YVarName = "Average";
            break;
        case DISTANCE: YVarName = "Distance";
            break;
        case RELERR: YVarName = "Relative Error";
            break;
        case VARIANCE: YVarName = "Variance";
            break;
        case STDDEV: YVarName = "Standard Deviation";
            break;
        case RESIDUAL: YVarName = "Residual";
            break;
        case KULLBACK: YVarName = "KL Distance";
            break;
    }
    for (std::deque<eval_execution>::iterator it = execs.begin();
            it != execs.end(); ee_index++, it++) {
        if ((selected[ee_index]) && ((*it).hasGlobal()) && ((globalYVar != RESIDUAL) || ((*it).hasResidual())) && ((globalYVar != KULLBACK) || ((*it).hasKullback()))) {
            line_real_set_t LRS(userMinXVal, userMaxXVal, userMinYVal, userMaxYVal);
            QPixmap* myPixmapP = new QPixmap(PMWidth, PMHeight);
            myPixmapP->fill(Qt::white);
            QPainter* myPainterP = new QPainter;
            myPainterP->begin(myPixmapP);
            // do the thing with the plot
            if (phanval > -0.5) { // if no error getting stuff from phan
                line_real_t phantomline("phantom", LRS.getMinX(), LRS.getMaxX(), phanval);
                LRS.add(phantomline);
            }
            it->addGlobVarLines(LRS, globalYVar);
            plot_t myPlotter(LRS); // now combine LRS with myPainterP
            QString evalexecname = it->getName();
            QString fullname = evalexecname;
            fullname += " :: ";
            fullname += YVarName;
            myPlotter.setName(fullname);
            myPlotter.setYAxisName(YVarName);
            QString XAxisName = "Iteration"; // as in all evals
            myPlotter.setXAxisName(XAxisName);
            // jk 2/2/2009 adding grayScale option for graphs
            if (isGrayScale) {
                myPlotter.plotGray(*myPainterP);
            } else {
                myPlotter.plot(*myPainterP);
            }
            x += 20;
            y += 20;
            displaylineswindow *dwp = new displaylineswindow(parent, fullname, myPixmapP, false, 0, x, y);
            dwp->show(); // and display!
        }
    } // --for
} // --eval_execution_set::spawnAllPlotsGlob()

void eval_execution_set::spawnAllPlotsPBP(QWidget* parent, unsigned PBPYVar,
        int userMinXVal, int userMaxXVal, double userMinYVal,
        double userMaxYVal, bool isGrayScale) {
    ///  std::cout << "in eval_execution_set::spawnAllPlotsPBP()" << std::endl;
    //  int PMWidth  = 650; // pixmap width, etc
    //  sd_line_t::WIDTH  = 650;
    int PMWidth = sd_line_t::WIDTH;
    //  int PMHeight = 550;
    //  sd_line_t::HEIGHT = 550;
    int PMHeight = sd_line_t::HEIGHT;
    unsigned ee_index = 0;
    int x = parent->pos().x();
    int y = parent->pos().y();
    QString YVarName = "e(";
    QString PBPYVarAsQString;
    PBPYVarAsQString.setNum(PBPYVar);
    YVarName += PBPYVarAsQString;
    YVarName += (')');
    for (std::deque<eval_execution>::iterator it = execs.begin();
            it != execs.end(); ee_index++, it++) {
        if ((selected[ee_index]) && ((*it).hasPBP())) {
            line_real_set_t LRS(userMinXVal, userMaxXVal, userMinYVal, userMaxYVal);
            QPixmap* myPixmapP = new QPixmap(PMWidth, PMHeight);
            myPixmapP->fill(Qt::white);
            QPainter* myPainterP = new QPainter;
            myPainterP->begin(myPixmapP);
            // do the thing with the plot: NO phantom for PBP
            it->addPBPVarLines(LRS, PBPYVar);
            plot_t myPlotter(LRS); // now combine LRS with myPainterP
            // can give name here based on (*it)
            QString evalexecname = it->getName();
            QString fullname = evalexecname;
            fullname += " :: ";
            fullname += YVarName;
            myPlotter.setName(fullname);
            QString xaxname = "Iteration";
            myPlotter.setXAxisName(xaxname);
            myPlotter.setYAxisName(YVarName);
            // jk 2/2/2009 adding grayScale option for graphs
            if (isGrayScale) {
                myPlotter.plotGray(*myPainterP);
            } else {
                myPlotter.plot(*myPainterP);
            }

            x += 20;
            y += 20;
            displaylineswindow *dwp = new displaylineswindow(parent, fullname, myPixmapP, false, 0, x, y);
            dwp->show(); // and display!
        }
    } // --for
} // --eval_execution_set::spawnAllPlotsPBP()

/** eval_execution_set::clear() !!!! erases all information!!! (to allow callee to get on with life) */
void eval_execution_set::clear() {
    ///  std::cout << "in eval_execution_set::clear()!!!" << std::endl;
    completed = extrema_calculated = false;
    same_x_forall = same_y_forall = false; // initialize to use "locally" calculated domain/range bounds
    selectionHasResidualvar = false;
    selectionHasKullbackvar = false;
    num_Es = 0;
    if (phantomeasp) delete phantomeasp;
    phantomeasp = 0;
    PBP_minima.clear();
    PBP_maxima.clear();
    selected.clear();
    execs.clear();
    minResidual = maxResidual = -1;
    minKullback = maxKullback = -1;
} // --eval_execution_set()

eval_execution_set::~eval_execution_set() {
    //  std::cout << "in eval_execution_set destructor!!!" << std::endl;
    if (phantomeasp) {
        delete phantomeasp; // and make sure no one else does this!
    }
} // --eval_execution_set dtor
//// // --class eval_execution_set

unsigned globmeas::NEXT_GMID = 0; // class variable initialization
unsigned globmeaset::NEXT_GMSID = 0; // class variable initialization
unsigned PBPmeas::NEXT_PMID = 0; // class variable initialization
unsigned PBPmeaset::NEXT_PMSID = 0; // class variable initialization
unsigned eval_execution::NEXT_EEID = 0; // class variable initialization

// --eval_helper.cpp
