/** @file line_window_onevar_t.cpp
    @author Deniz
    @description implementation of functions in line_window_onevar_t
    licensed under (open-source) QPL v1.0
    which accompanies this distribution in the file QPL
*/

#include <iostream>
#include <list>
#include <stdexcept>
#include <exception>

#include "verbosity.hpp"
#include "sd_line_t.hpp"
#include "line_real_t.hpp"
#include "line_window_onevar_t.hpp"

typedef std::list<line_window_t> data_onevar_t;

/** class line_window_onevar_t 
    corresponds to a set of maybe disconnected (due to falling out of window boundaries)
    segments for a single variable, due to single execution
    All these hacks are due to a bug i found in QT API (both versions)
    when plotting big numbers: now being superparanoid to drawLine() ONLY within the window, 
    which is not necessarily as precise but at least guarantees that 
    there will be lines when there should be. 
*/

/** constuctor/1 takes one line with real coords, potentially makes many point sets from it    *
 *  pre: line_real has at least 1 point                                                        *
 *  pre: x values in line_real are strictly increasing (as is the case in snark14 eval output) *
 *  (as well as in column/row profiles?) i believe those are not even given as (x,y)  (yet)    *
 *  trusts Painter::drawLine() with between interior and JUST outside the paint window.        *
 *  this is where conversion from reals to window coords actually happens                      *
 *  2004 version WAS TOO SIMPLE MINDED.                                                        *
 *    even if both points are out of range, we still want to see a line,                       *
 *        without there being an inside-the-box first point!                                   *
 */
line_window_onevar_t::line_window_onevar_t(line_real_t& line_real) {
  if(std::verbose > 2) {
    std::cerr << std::endl << "line_window_onevar_t ctor++" << std::endl;
  }
  if(line_real.empty()) {
    std::cerr << ("Empty line passed to line_window_onevar_t constructor/3");
    throw std::runtime_error("Empty line passed to line_window_onevar_t constructor/3");
  }
  //    XWINPAD and YWINPAD are defined consts (int)
  if ((3*sd_line_t::XWINPAD > sd_line_t::WIDTH) || (3*sd_line_t::YWINPAD > sd_line_t::HEIGHT)) {
    std::cerr << "PADDING WITHIN WINDOW IS HUGE!  ADJUST XWINPAD / YWINPAD, codeur!" << std::endl;
    throw std::logic_error("in line_window_onevar_t ctor, too huge padding wrt W/H");
    // this is why it's a bad idea to let the user set WIDTH and HEIGHT
  }
  plotname = line_real.getName();
  int xrmin = line_real.getMinX(); // "x real min" etc.  real as opposed to window coords
  int xrmax = line_real.getMaxX();
  double yrmin = line_real.getMinY();
  double yrmax = line_real.getMaxY();
  // perform fudging defensively
  if(0 < sd_line_t::X_HALF_FUDGE_INCREMENT) {
    xrmin -= sd_line_t::X_HALF_FUDGE_INCREMENT;
    xrmax += sd_line_t::X_HALF_FUDGE_INCREMENT;
  }
  int xr_range = xrmax - xrmin; // underscore between r's for emphasis
  double yr_range = yrmax - yrmin;
  if(std::verbose > 2) {
    std::cerr << "xr_range: " << xr_range << ", yr_range: " << yr_range << std::endl;
  }
  if(0 < sd_line_t::Y_HALF_FUDGE_FACTOR) {
    double y_half_fudge = yr_range * sd_line_t::Y_HALF_FUDGE_FACTOR;
    yrmin -= y_half_fudge;
    yrmax += y_half_fudge;
    yr_range = yrmax - yrmin; // calculate post-fudge yr_range
  }
  // convert to lattice points of [XWINPAD, WIDTH-XWINPAD) x [YWINPAD, HEIGHT-YWINPAD)
  // NB: y reverses direction
  int cr_x, pr_x;    // current and previous real x's
  double cr_y, pr_y; // current and previous real y's
  int xwmin = sd_line_t::XWINPAD;
  //      int xwmax = WIDTH - 1 - XWINPAD;
  int xwrange = sd_line_t::WIDTH - 1 - 2*sd_line_t::XWINPAD;
  int ywmin = sd_line_t::YWINPAD;
  // the following can be very slightly out of range: trust Qt to be OK with it
  int ywmax = sd_line_t::HEIGHT - 1 - sd_line_t::YWINPAD;
  int ywrange = ywmax - ywmin;
  double yscalar = static_cast<double>(ywrange) / yr_range;
  double xscalar = static_cast<double>(xwrange) / static_cast<double>(xr_range);
  line_window_t LW; // treated like scratch paper: fills up, pushed, emptied, filled again...
  int xw, yw;
  double xr;  // , yr;
  // xwrange corresponds to xr_range
  // ywrange corresponds to yr_range
  // the line (xr = xrmin) intersects the  *left side*  of (pixels with "x coordinate" xwmin)
  // the line (xr = xrmax) intersects the *right side*  of (pixels with "x coordinate" xwmax)
  // the line (yr = yrmin) intersects the *bottom side* of (pixels with "y coordinate" ywmax)
  // the line (yr = yrmax) intersects the  *top side*   of (pixels with "y coordinate" ywmax)
  // in general:
  // xw = xwmin + static_cast<int> (xscalar * (xr - xrmin));
  // yw = ywmin + static_cast<int> (yscalar * (yrmax - yr));
  bool firstdatapoint = true;
  bool firstpointwithinxrange = false;
  bool prev_r_low, prev_r_high, cur_r_low, cur_r_high;
  line_real_t::data_real_t::const_iterator it = line_real.data.begin();
  if(it == line_real.data.end()) {
    std::cerr << "line_window_onevar_t ctor: No data points... but checked for this already?" << std::endl;
    return;
  }
  // first real point gets 'special treatment' in that it has no previous data point
  cr_x = it->x;
  if(cr_x > xrmax) return; // nothing will be within range, x's increasing
  cr_y = it->y;
  cur_r_low  = (cr_y < yrmin);
  cur_r_high = (cr_y > yrmax);
  // for loop will execute only as long as cur x is to the left
  while((it != line_real.data.end()) && (cr_x < xrmin)) {
    it++;
    cr_x = it->x;
    cr_y = it->y;
    cur_r_low  = (cr_y < yrmin);
    cur_r_high = (cr_y > yrmax);
  }
  if(cr_x < xrmin) {
    if(std::verbose > 1) {
      std::cerr << "line_window_onevar_t ctor returning because everything was to the left of x range" << std::endl;
    }
    return; // nothing was within range
  }
  // since we started with less than x (otherwise would have returned)
  // now cur x is within the range, since x range contains at least one x value!
  if(cr_x > xrmax) {
    if(std::verbose > 1) {
      std::cerr << "in line_window_onevar_t ctor: But this is impossible!" << std::endl;
    }
    return;
  }
  // now, we have the first data point whose real x is within range of interest
  if(! (cur_r_low || cur_r_high)) {
    xw = xwmin + static_cast<int> (xscalar * (cr_x - xrmin));
    yw = ywmin + static_cast<int> (yscalar * (yrmax - cr_y));
    LW.add(xw, yw);
  }
  it++; // now do all the other points!
  for(; it != line_real.data.end(); it++) {
    pr_x = cr_x;
    pr_y = cr_y;
    prev_r_low = cur_r_low;
    prev_r_high = cur_r_high;
    cr_x = it->x;
    if(cr_x > xrmax) break; // this one is fine: we won't look at the rest of the data...
    cr_y = it->y;
    cur_r_low  = (cr_y < yrmin);
    cur_r_high = (cr_y > yrmax);
    // from now on, lowness and highness only concerns y, until we run out, which is the end
    // in general: 9 (3*3) possibilities
    if(prev_r_low) { // 1
      if(cur_r_low) { // 1.1
	// do nothing! (except to set prev := cur, which happens near the closing brace)
      } else if(cur_r_high) { // 1.2
	// calculate two x-intercepts in between prev and cur... first the lower one
	// "xr_intercept_low" is the x intercept of yrmin
	double xr_intercept_low = (cr_x) - ((cr_y - yrmin ) * (cr_x - pr_x)) / (cr_y - pr_y);
	xw = xwmin + static_cast<int> (xscalar * (xr_intercept_low - xrmin));
	LW.add(xw,ywmax); // ywmax <=> yrmin
	// "xr_intercept_high" means the x intercept with yrmax
	double xr_intercept_high = (cr_x) - ((cr_y - yrmax ) * (cr_x - pr_x)) / (cr_y - pr_y);
	xw = xwmin + static_cast<int> (xscalar * (xr_intercept_high - xrmin));	    
	LW.add(xw,ywmin); // ywmin <=> yrmax
	data.push_back(LW);  // flush
	LW.clear();
      } else { // 1.3
	// prev was low, cur gets us back in the game
	// add x-intercept first, then cur.
	double xr_intercept_low = (cr_x) - ((cr_y - yrmin ) * (cr_x - pr_x)) / (cr_y - pr_y);
	xw = xwmin + static_cast<int> (xscalar * (xr_intercept_low - xrmin));
	LW.add(xw,ywmax); // ywmax <=> yrmin
	// convert (cr_x, cr_y) to window coords
	xw = xwmin + static_cast<int> (xscalar * (cr_x - xrmin));
	yw = ywmin + static_cast<int> (yscalar * (yrmax - cr_y));
	LW.add(xw,yw);
      }
    } else if(prev_r_high) { // 2
      if(cur_r_low) { // 2.1
	// calculate two x-intercepts in between prev and cur
	double xr_intercept_high = (cr_x) - ((cr_y - yrmax ) * (cr_x - pr_x)) / (cr_y - pr_y);
	xw = xwmin + static_cast<int> (xscalar * (xr_intercept_high - xrmin));	    
	LW.add(xw,ywmin); // ywmin is top of window
	double xr_intercept_low = (cr_x) - ((cr_y - yrmin ) * (cr_x - pr_x)) / (cr_y - pr_y);
	xw = xwmin + static_cast<int> (xscalar * (xr_intercept_low - xrmin));
	LW.add(xw,ywmax); // ywmax <=> yrmin
	data.push_back(LW);  // flush
	LW.clear();
      } else if(cur_r_high) { // 2.2
	// do nothing! (except to set prev eq to cur, which happens near the closing brace)
      } else { // 2.3
	// prev was high, cur gets us back in the game
	// add x-intercept first, then cur
	double xr_intercept_high = (cr_x) - ((cr_y - yrmax ) * (cr_x - pr_x)) / (cr_y - pr_y);
	xw = xwmin + static_cast<int> (xscalar * (xr_intercept_high - xrmin));
	LW.add(xw,ywmin); // ywmin <=> yrmax
	xw = xwmin + static_cast<int> (xscalar * (cr_x - xrmin));
	yw = ywmin + static_cast<int> (yscalar * (yrmax - cr_y));
	LW.add(xw,yw);
      }
    } else { // 3
      if(cur_r_low) { // 3.1
	// prev was OK: add new x-intercept (other point is yrmin aka ywmax) and flush
	double xr_intercept_low = (cr_x) - ((cr_y - yrmin ) * (cr_x - pr_x)) / (cr_y - pr_y);
	xr = xr_intercept_low;
	xw = xwmin + static_cast<int> (xscalar * (xr - xrmin));
	LW.add(xw,ywmax); // ywmax is bottom of window
	data.push_back(LW);  // flush
	LW.clear();
      } else if(cur_r_high) { // 3.2
	// prev was OK: add new x-intercept (other point is yrmax aka ywin)	and flush
	double xr_intercept_high = (cr_x) - ((cr_y - yrmax ) * (cr_x - pr_x)) / (cr_y - pr_y);
	xw = xwmin + static_cast<int> (xscalar * (xr_intercept_high - xrmin));
	LW.add(xw,ywmin); // ywmin <=> yrmax
	data.push_back(LW);  // flush
	LW.clear();
      } else { // 3.3
	// simplicity itself: prev and cur both ok, just add cur to LW
	xw = xwmin + static_cast<int> (xscalar * (cr_x - xrmin));
	yw = ywmin + static_cast<int> (yscalar * (yrmax - cr_y));
	LW.add(xw,yw);
      }
    }
  } // --for
  // flush if LW is non-empty
  if(!LW.empty()) {
    data.push_back(LW);
  }
  else {
    /// std::cout << "WARNING: nothing generated by line_window_onevar_t ctor" << std::endl;
    /// commented out because this can and does happen and it's ok.
  }
  if(std::verbose > 2) {
    std::cerr << std::endl << "line_window_onevar_t ctor--" << std::endl;
  }
} // -- line_window_onevar_t constructor, might just be the longest constructor ever written


bool line_window_onevar_t::empty() { return data.empty(); }  // --line_window_onevar_t::empty()


QString line_window_onevar_t::getName() { return plotname; } // --line_window_onevar_t::getName()


//    void add(line_window_t& LW) {
//      data.push_back(LW);
//    } // --line_window_onevar_t::add()


void line_window_onevar_t::plot(QPainter& P, QColor& C, Qt::PenStyle & Ps) {
  for(data_onevar_t::iterator it = data.begin();
      it != data.end(); it++) {
    it->plot(P, C, Ps);
  }
} // --line_window_onevar_t::plot()


void line_window_onevar_t::show() {
  for(data_onevar_t::iterator it = data.begin();
      it != data.end(); it++) {
    it->show();
  }
  std::cerr << std::endl;
}
