/** @file picture.cpp
    @package snark14Input
    @author Bruno M. Carvalho and Deniz Sarioz
    licensed under (open-source) QPL v1.0
    which accompanies this distribution in the file QPL
*/

#include "picture.h"
#include "variables.h"
#include "misc.h"

#include <qbuttongroup.h>
#include <qframe.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qpushbutton.h>
#include <qradiobutton.h>
#include <qlayout.h>
#include <qvariant.h>
#include <qtooltip.h>
#include <qwhatsthis.h>
#include <qmessagebox.h>

/** 
 *  Constructs a picturewindow which is a child of 'parent', with the 
 *  name 'name' and widget flags set to 'f' 
 *
 *  The dialog will by default be modeless, unless you set 'modal' to
 *  TRUE to construct a modal dialog.
@param void
@author Bruno M. Carvalho
@version 1.0 */
picturewindow::picturewindow( QWidget* parent,  const char* name, bool modal, WFlags fl )
    : QDialog( parent, name, modal, fl )
{
    if ( !name )
	setName( "picturewindow" );
    resize( 400, 250 ); 
    setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)0, (QSizePolicy::SizeType)0, sizePolicy().hasHeightForWidth() ) );
    setMinimumSize( QSize( 400, 250 ) );
    setMaximumSize( QSize( 400, 250 ) );
    setCaption( tr( "Picture Window" ) );

    picturelabel = new QLabel( this, "picturelabel" );
    picturelabel->setGeometry( QRect( 10, 10, 100, 30 ) ); 
    picturelabel->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)0, (QSizePolicy::SizeType)0, picturelabel->sizePolicy().hasHeightForWidth() ) );
    picturelabel->setMinimumSize( QSize( 100, 30 ) );
    picturelabel->setMaximumSize( QSize( 100, 30 ) );
    picturelabel->setText( tr( "PICTURE" ) );

    picturebuttongroup = new QButtonGroup( this, "picturebuttongroup" );
    picturebuttongroup->setGeometry( QRect( 5, 35, 385, 130 ) ); 
    picturebuttongroup->setTitle( tr( "type" ) );
    picturebuttongroup->setAlignment( int( QButtonGroup::AlignHCenter ) );

    reconstructionbutton = new QRadioButton( picturebuttongroup, "reconstructionbutton" );
    reconstructionbutton->setGeometry( QRect( 10, 40, 180, 24 ) ); 
    reconstructionbutton->setText( tr( "RECONSTRUCTION" ) );
    QObject::connect(reconstructionbutton,SIGNAL(clicked()),this,SLOT(activateReconstructionFields()));

    testbutton = new QRadioButton( picturebuttongroup, "testbutton" );
    testbutton->setGeometry( QRect( 10, 80, 65, 24 ) ); 
    testbutton->setText( tr( "TEST" ) );
    QObject::connect(testbutton,SIGNAL(clicked()),this,SLOT(deactivateReconstructionFields()));

    nelem = new QLineEdit( picturebuttongroup, "nelem" );
    nelem->setGeometry( QRect( 200, 35, 80, 30 ) ); 
    nelem->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)0, (QSizePolicy::SizeType)0, nelem->sizePolicy().hasHeightForWidth() ) );
    nelem->setEnabled(FALSE);

    nelemlabel = new QLabel( picturebuttongroup, "nelemlabel" );
    nelemlabel->setGeometry( QRect( 220, 15, 50, 20 ) ); 
    nelemlabel->setText( tr( "nelem" ) );
    nelemlabel->setEnabled(FALSE);

    pixelsize = new QLineEdit( picturebuttongroup, "pixelsize" );
    pixelsize->setGeometry( QRect( 295, 35, 80, 30 ) ); 
    pixelsize->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)0, (QSizePolicy::SizeType)0, pixelsize->sizePolicy().hasHeightForWidth() ) );
    pixelsize->setEnabled(FALSE);

    pixelsizelabel = new QLabel( picturebuttongroup, "pixelsizelabel" );
    pixelsizelabel->setGeometry( QRect( 305, 15, 70, 20 ) ); 
    pixelsizelabel->setSizePolicy( QSizePolicy( (QSizePolicy::SizeType)0, (QSizePolicy::SizeType)0, pixelsizelabel->sizePolicy().hasHeightForWidth() ) );
    pixelsizelabel->setMinimumSize( QSize( 70, 20 ) );
    pixelsizelabel->setMaximumSize( QSize( 70, 20 ) );
    pixelsizelabel->setText( tr( "pixel-size" ) );
    pixelsizelabel->setEnabled(FALSE);

    Line10 = new QFrame( this, "Line10" );
    Line10->setGeometry( QRect( 0, 175, 420, 16 ) ); 
    Line10->setFrameStyle( QFrame::HLine | QFrame::Sunken );

    okbutton = new QPushButton( this, "okbutton" );
    okbutton->setGeometry( QRect( 50, 195, 100, 40 ) ); 
    okbutton->setText( tr( "OK" ) );
    QObject::connect(okbutton,SIGNAL(clicked()),this,SLOT(checkValues()));

    cancelbutton = new QPushButton( this, "cancelbutton" );
    cancelbutton->setGeometry( QRect( 250, 195, 100, 40 ) ); 
    cancelbutton->setText( tr( "Cancel" ) );
    QObject::connect(cancelbutton,SIGNAL(clicked()),this,SLOT(reject()));
}

/**  
 *  Destroys the object and frees any allocated resources
 */
picturewindow::~picturewindow()
{
    // no need to delete child widgets, Qt does it all for us
}

/** Activates type fields 
@param void
@author Bruno M. Carvalho
@version 1.0 */
/* void picturewindow::activateFields()
{
  if(picturebutton->isChecked()) {
    reconstructionbutton->setEnabled(TRUE);
    testbutton->setEnabled(TRUE);
    picturebuttongroup->setEnabled(TRUE);
  }
  else {
    picturebuttongroup->setEnabled(FALSE);
    reconstructionbutton->setEnabled(FALSE);
    testbutton->setEnabled(FALSE);
    nelem->setEnabled(FALSE);
    pixelsize->setEnabled(FALSE);
    nelemlabel->setEnabled(FALSE);
    pixelsizelabel->setEnabled(FALSE);
  }
} */

/** Activates reconstruction fields 
@param void
@author Bruno M. Carvalho
@version 1.0 */
void picturewindow::activateReconstructionFields()
{
  if(reconstructionbutton->isChecked()) {
    nelem->setEnabled(TRUE);
    pixelsize->setEnabled(TRUE);
    nelemlabel->setEnabled(TRUE);
    pixelsizelabel->setEnabled(TRUE);
  }
  else {
    nelem->setEnabled(FALSE);
    pixelsize->setEnabled(FALSE);
    nelemlabel->setEnabled(FALSE);
    pixelsizelabel->setEnabled(FALSE);
  }
}

/** Deactivates reconstruction fields 
@param void
@author Bruno M. Carvalho
@version 1.0 */
void picturewindow::deactivateReconstructionFields()
{
  if(testbutton->isChecked()) {
    nelem->setEnabled(FALSE);
    pixelsize->setEnabled(FALSE);
    nelemlabel->setEnabled(FALSE);
    pixelsizelabel->setEnabled(FALSE);
  }
}

/** Returns a string for PICTUE input sequence
@param void
@return picture
@author Bruno M. Carvalho
@version 1.0 */
QString picturewindow::getOutput()
{
  QString output;

  output.sprintf("PICTURE ");
  if(reconstructionbutton->isChecked()) {
    output.append("RECONSTRUCTION ");
    output+=nelem->text();
    output.append(" ");
    output+=pixelsize->text();
    output.append(" ");
  }
  if(testbutton->isChecked()) 
    output.append("TEST");
  //output.append("\n");

  return output;
}


void picturewindow::checkValues()
{
  int f=0,n=0;
  float p=0;

  if(reconstructionbutton->isChecked()) {
    n=strtoi((const char *)nelem->text());
    if(errnum) 
      f=1;
    if(n<1 || !(n%2))
      f=2;
    p=strtof((const char *)pixelsize->text());
    if(errnum) 
      f=3;
    if(p<=0)
      f=4;
  }
  switch(f) {
  case 0: accept();
    break;
  case 1: QMessageBox::information(this,"SnarkInput","Error!\n"
                                   "Invalid nelem format.");
          nelem->setFocus();
    break;
  case 2: QMessageBox::information(this,"SnarkInput","Error!\n"
                                   "nelem must be positive and odd.");
          nelem->setFocus();
    break;
  case 3: QMessageBox::information(this,"SnarkInput","Error!\n"
                                   "Invalid pixel-size format.");
          pixelsize->setFocus();
    break;
  case 4: QMessageBox::information(this,"SnarkInput","Error!\n"
                                   "pixel-size must be greater than 0.");
          pixelsize->setFocus();
    break;
  }
}
